/**************************************************
 * RCSId: $Id$
 *
 * HRI record functions
 * Project: WSAT
 * Author: R. Alblas
 *
 * History: 
 * $Log$
 **************************************************/
/*******************************************************************
 * Copyright (C) 2000 R. Alblas. 
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software. If not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 ********************************************************************/
#ifdef __ADD_PAR__ 

#include <unistd.h>
#include <malloc.h>
#include "pp_dec_interf.h"
#include "wsat.h"
#include "menu_wsat.h"
#if _HRI1_
#  include "hri1.h"
#endif

extern PREFS prefs;


/* ====== Define button text for HRI Record window ====== */
#define LAB_DEC       "Decoder"
#define LAB_GEN       "Generator"
#define LAB_RECALL    "Record all"
#define LAB_Start     "Start/Stop"
#define LAB_Stop      "Stop"
#define LAB_Save      "Save"

#define LAB_ADVANCED  "Advanced"
#define LAB_ADV_SET   "!advanced"
#define LAB_Exit      " Exit "
#define Lab_Hriwnd    "Record HRI"
#define LAB_Recording "Recording"
#define LAB_RECLINES  "Line #"
#define LAB_Satname   "Sat name "
#define LAB_Encryp    "Encrypted:"
#define LAB_Cntrec    "Recorded"
#define LAB_Cntencr   "Encrypted"

static GtkWidget *hide_but;

/*************************************
 * Get a HRI-frame from decoder.
 * Decoder is connected to parallel port.
 * Return:
 *   0 -> Abort by user
 *   1 -> OK; decoder in-sync
 *   2 -> OK, but decoder out-sync
 *************************************/
int getpp_hri_frame(guint16    *frame,        /* interf. w. rest of funcs */
                    guint8     *frame8,       /* for easy header interpret. */
                    RECORDINFO *rdi)
{
  int datstat;
  int words_remaining;
  gboolean start_frame_flag=FALSE;
  int cnt_nosig=0;

  words_remaining=rdi->fri.words_per_frame;
  while (words_remaining > 0)              /* Read all words in a frame */
  {
/* Get data from parallel port */
    if (!(get_pp_data(rdi,&datstat)))
    {
      return 0;
    }

/* Check if received word contains a frame sync pulse from decoder.
   If sync pulse received, this is the first word of a frame.
   (Normally, this is the case each time getpp_hri_frame is entered.)
*/
    if (STARTFRAME_PP(datstat))
    {
/* Now,  enable to save remaining words of frame. */
      start_frame_flag=TRUE;
    }

/* Save remaining words of frame. */
    if (start_frame_flag)
    {
      if (frame)
      {
        *frame=datstat&0xff;
        frame++;
      }
      if (frame8)
      {
        *frame8=datstat&0xff;
        frame8++;
      }
      words_remaining--;
      cnt_nosig=0;
    }
    else
    {
/* If no readyn: make sure that now and then gui is activated */
      cnt_nosig++;
      if (cnt_nosig>1000)
      {
        while (g_main_iteration(FALSE));
        if (!rdi->recording) break;
        cnt_nosig=0;
      }
    }
  }
  if (!INSYNC_PP(datstat)) return 2;        /* decoder not in-sync. */
  return 1;                                 /* decoder in-sync. */
}


/*************************************
 * Read from decoder until some picture characteristics are known.
 * Determine frame type and encryption.
 * Error-resistancy is built-in by means of majority voting mechanism.
 *************************************/
#define THRESHOLD 3             /* threshold for majority voting */
gboolean catch_hri_start(PICINFO *pci,RECORDINFO *rdi)
{
  gboolean found=FALSE;
  guint8 *frame8;
  HRI_SUBFRHDR *hri_sf;
  int fpsf8=0,fpsf4=0,encrypted=0;
  guint16 satind=0xffff;
  guint16 frmcnt;

/* allocate mem to hold 1 frame */
  frame8=(guint8 *)malloc(rdi->fri.words_per_frame*sizeof(guint8));
  hri_sf=(HRI_SUBFRHDR *)frame8;

/* Store some default numbers */
  pci->hri_pci.frm_p_subfrm=4;
  pci->hri_pci.encrypted=FALSE;

/* Start searching. */
  while (getpp_hri_frame(NULL,frame8,rdi))
  {
    while (g_main_iteration(FALSE));
    if (!rdi->recording) break;

/* Needed first frame of subframe. */
    if (!(hri_sf->id&0x0f))                     /* first frame? */
    {
      frmcnt=GUINT16_FROM_BE(hri_sf->cur_subfr);
      if (frmcnt==1) 
        satind=GUINT16_FROM_BE(hri_sf->sat_ind);

      if ((hri_sf->text_ind&0x0f)!=0x00)  /* 22 */
        encrypted++;
      else
       if (encrypted>0)  encrypted--;

      if ((hri_sf->id&0xf0)==0x70)
        fpsf8++;
      else
        if (fpsf8>0) fpsf8--;

      if ((hri_sf->id&0xf0)==0x30)
        fpsf4++;
      else
        if (fpsf4>0) fpsf4--;

      if ((fpsf8>THRESHOLD) || (fpsf4>THRESHOLD))
        break;
    }
  }

/* Determine type of subframe */
  if (fpsf8>THRESHOLD) 
  {
    pci->hri_pci.frm_p_subfrm=8;
    found=TRUE;
  }
  else if (fpsf4>THRESHOLD) 
  {
    pci->hri_pci.frm_p_subfrm=4;
    found=TRUE;
  }

/* Determine encryption */
  if (encrypted>0) 
    pci->hri_pci.encrypted=TRUE;

/* determine sat-name */
  hri_satind2satname(satind,pci->hri_pci.satname);

  free(frame8);
  return found;
}

/*************************************
 * Record one HRI picture.
 *************************************/
void Recordpp_one_Hri(GtkWidget  *hri_window,
                      GtkWidget  *main_window,
                      RECORDINFO *rdi,
                      PICINFO    *pci,
                      VIEWINFO   *vwi)
{
  RGBI *rgbi=Get_RGBI(main_window);
  GtkWidget *drawing_area=Find_Widget(main_window,"GTK_DRAWING_AREA");
  int xdrawmax=0;
  int ydrawmax=0;
  int ydraw=0,y=0;
  int nr_frames;
  guint16 *frame;
  guint16 *frame2;
  int ox,nr_fr=0,i;
  int out_sync_counter=0;
  gboolean run_record=TRUE;
  int status,prev_status=-1;
  int ypos;

  guint *xmap=NULL;
  guint *ymap=NULL;
  guint *zmap=NULL;

  if (drawing_area)
  {
    xdrawmax=drawing_area->allocation.width;
    ydrawmax=drawing_area->allocation.height;
  }

/* If a file was open for read, close it */
  close_picfile(pci);
  
/* open file if we want to save the record */
  if (rdi->save_flag)
  {
    if (!(open_new_recfile(rdi->file_prefix,rdi->file_suffix,rdi)))
    {
      Create_Message("Error","Can't open file %s!",rdi->file_name);
      return;
    }
    Set_Entry(hri_window,LAB_Recording,"%s",rdi->file_name);
    if (rdi->batch_mode) log_message("Start recording in %s.",rdi->file_name);

#if _HRI1_
    write_hri1_hdr(rdi,0);
#endif
  }
  else
    Set_Entry(hri_window,LAB_Recording,"(no save)");


/* allocate mem to hold 1 frame */
  frame=(guint16 *)malloc(rdi->fri.words_per_frame*sizeof(guint16));


/*
  Determine some pic-parameters for real-time show, such that 1 picture 
  fills exactly 1 screen.
  . width is assumed to be #frames/line*360;
      - including the grid-info in the last part of the last frame
      - excluding 4 first header bytes in each frame
  . height is assumed to be same as width,
*/
  pci->clrmap.size=256;
  pci->width=360*pci->hri_pci.frm_p_subfrm;
  pci->height=ydrawmax;        /* force map_pic to do 1:1 mapping on screen */
  rdi->height=pci->width;
  if ((drawing_area) && (!rdi->height))
    rdi->height=drawing_area->allocation.height;
  pci->fri=calc_hriframeinfo();


/* Define some view characteristics for the channel to show real-time */
  vwi->zval=zoom_full();
  vwi->scan_hor=W_E;
  vwi->scan_ver=N_S;
  vwi->lum_min=0;
  vwi->lum_max=255;

/* Determine mapping of this channel.
   ymap from map_pic is translation pic_y -> screen_y, but here vice-versa 
   is needed. So, ymap can only be used to handle sat-direction.
   Actual y-mapping is determined by means of rdi->height.
*/
  map_pic(main_window,pci,vwi,FALSE,&xmap,&ymap,&zmap);

/* allocate mem for line to show real-time */
  frame2=(guint16 *)malloc(rdi->fri.words_per_frame*
                          pci->hri_pci.frm_p_subfrm*sizeof(guint16));

/* Start the record loop. Each loop is 1 frame. */
  nr_frames=0;
  while ((run_record) && (rdi->recording))
  {
    if (rdi->batch_mode)
    {
      if ((rdi->sel_hw) && (nr_frames>100)) rdi->recording=FALSE;
    }
    else
    {
      while (g_main_iteration(FALSE));
    }

/* Get a HRI-frame. Show if signal lost. */
    status=getpp_hri_frame(frame,NULL,rdi);
    if ((prev_status>=0) && (status!=prev_status))
    {
      if ((status) && (!prev_status))
        Set_Entry(hri_window,LAB_Recording,"Signal restored");
      else if (!status)
        Set_Entry(hri_window,LAB_Recording,"!! No signal!!");
    }
    prev_status=status;

/* Handle in-sync info decoder */
    if (status==2)                            /* out-sync */
      out_sync_counter++;
    else                                      /* in-sync */
      out_sync_counter=0;
    if (out_sync_counter > rdi->hri_stop_after_out_sync)
      break;                                  /* abort by decoder */

/* Synchronize frame number */
    if ((frame[3]&0x0f) < pci->hri_pci.frm_p_subfrm)
      nr_fr=frame[3]&0x0f;

/* determine offset current frame */
    ox=nr_fr*360;
    for (i=0; i<360; i++) frame2[ox+i]=frame[i+4];

    nr_fr++;
    if (nr_fr>=pci->hri_pci.frm_p_subfrm)
    {
      nr_fr=0;

      if (ymap)
      {
/* Do actual mapping */
        if (rdi->height) ypos=y*(ydrawmax-1)/rdi->height;

/* Restart at top if canvas is full */
        if (ypos>=ydrawmax) y=0;
        ydraw=*(ymap+ypos);

/* solve under/overflow (just in case; should never happen) */
        if (ydraw<0) ydraw=0;
        if (ydraw>=ydrawmax) ydraw=ydrawmax-1;

/* Draw the received line into the canvas. */
        draw_a_line_bw(frame2,xdrawmax,ydraw,xmap,zmap,255,rgbi,NULL);
        Refresh_Rect(drawing_area,0,ydraw,
                     drawing_area->allocation.width,
                     1);
        y++;
      }
      if (!(y%10)) Set_Entry(hri_window,LAB_RECLINES,"%d",y);
    }

/* Save frame */
    if (rdi->save_flag)
#if _HRI1_
      write_hri1_line(rdi,frame);
#endif

    nr_frames++;

  }

/* End of recording. */
  if (rdi->save_flag)
  {
    close_recfile(rdi);
  }

  free(frame);
  free(frame2);
  Set_Entry(hri_window,LAB_Recording,"Stop");
  if (rdi->batch_mode) log_message("Stop recording after %d frames.",nr_frames);

  return;
}


/*************************************
 * Record HRI.
 *************************************/
void Recordpp_Hri(GtkWidget *window,
                  RECORDINFO *rdi,
                  PICINFO *pci,
                  VIEWINFO *vwi)
{
  GtkWidget *mainwnd=Find_Window(window,Lab_Main);

  int addr_portd=rdi->parport_base;
  int count_rec=0,count_encr=0;
/* Set counters */
  Set_Entry(window,LAB_Cntrec,"%d",count_rec);
  Set_Entry(window,LAB_Cntencr,"%d",count_encr);

/* calculate some frame characteristics */
  rdi->fri.words_per_frame=LEN_HRIFRAME;
  rdi->fri.bytes_per_line=LEN_HRIFRAME;
  strcpy(rdi->file_prefix,rdi->file_hribase);
  strcpy(rdi->file_suffix,rdi->file_hriext);

/* Check if it is allowed to access the parallel port */
  if (ioperm(addr_portd,3,1))
  {
    Create_Message("Error",PAR_ERR_MESSAGE,addr_portd);
    return;
  }

/* Start the record loop; Each loop is 1 picture. */
  while (rdi->recording)
  {
/* Wait until the decoder is in-sync for some time */
    Set_Entry(window,LAB_Recording,"Wait");
    if (rdi->batch_mode) log_message("Wait for in-sync.");
    if (!(wait_for_insync(rdi,rdi->hri_start_after_in_sync))) break;


/* Read a few frames to determine format. */
    if (!(catch_hri_start(pci,rdi))) return;
    Set_Entry(window,LAB_Satname,pci->hri_pci.satname);

/* Abort if encrypted; start record if not */
    if (pci->hri_pci.encrypted)
    {
/* Encrypted; Wait until the decoder is out-sync for some time. */
/*        Or: Start record. */
      if (rdi->batch_mode) log_message("Encrypted.");
      count_encr++;
      Set_Entry(window,LAB_Cntencr,"%d",count_encr);
      Set_Entry(window,LAB_Encryp,"Y");
      if (rdi->record_all)
      {
        Recordpp_one_Hri(window,mainwnd,rdi,pci,vwi);
      }
      else
      {
        if (rdi->batch_mode) log_message("Wait for out-sync.");
        if (!(wait_for_outsync(rdi,rdi->hri_stop_after_out_sync))) break;
      }
    }
    else
    {
/* Not encrypted; start record. */
      count_rec++;
      Set_Entry(window,LAB_Encryp,"N");
      Set_Entry(window,LAB_Cntrec,"%d",count_rec);
      Recordpp_one_Hri(window,mainwnd,rdi,pci,vwi);
    }

/* If just 1 record then close the program (record finished) */
    if (rdi->one_record)
    {
      rdi->recording=FALSE;
      if (!rdi->batch_mode)
        gtk_main_quit();
    }
  }
  Set_Entry(window,LAB_Recording,"Off");

/* Open file if recorded */
  if ((!rdi->batch_mode) && (rdi->save_flag))
  {
    strcpy(pci->file_name,rdi->file_name);
    open_file(pci,vwi,mainwnd);
  }
  if (rdi->batch_mode) log_message("End recording.");
}



/*************************************
 * Get button signals for Record functions
 *************************************/
void Record_Hri_Commands(GtkWidget *widget, gpointer data)
{
  GtkWidget *hri_window=gtk_widget_get_toplevel(widget);
  GtkWidget *main_window=Find_Window(hri_window,Lab_Main);
  PICINFO *pci=Get_Picinfo_db(widget);
  VIEWINFO *vwi=Get_Viewinfo_db(widget);
  RECORDINFO *rdi=Get_Recordinfo_db(widget);

  int addr_portd=rdi->parport_base;
  char *name;
  gboolean reprogramming_fpga=FALSE;

  name=(char *)data;

/* Decoder/generator choice */
  if (!strcmp(name,LAB_DEC))
  {
    rdi->gen1_dec0=(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))? FALSE : TRUE);
    reprogramming_fpga=TRUE;
  }

/* Recording to save or not to save... */
  if (!strcmp(name,LAB_Save))
  {
    rdi->save_flag=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
  }

  if (!strcmp(name,LAB_RECALL))
  {
    rdi->record_all=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
  }

  if (!strcmp(name,LAB_ADVANCED))
  {
    if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)))
    {
      hide_but=Find_Widget(hri_window,LAB_ADV_SET);
      if (hide_but) gtk_widget_hide(hide_but); 
    }
    else
    {
      if (hide_but) gtk_widget_show(hide_but);
      hide_but=NULL;
    }
  }


/* Start record */
  if ((!strcmp(name,LAB_Start)) && (Update_Togglelabel(widget)))
  {
    rdi->recording=TRUE;
/*
    Sense_Button(hri_window,LAB_Start,FALSE);
*/
    Sense_Button(hri_window,LAB_Exit,FALSE);
    Sense_Button(hri_window,LAB_Save,FALSE);
    Recordpp_Hri(hri_window,rdi,pci,vwi);
/*
    Sense_Button(hri_window,LAB_Start,TRUE);
*/
    Sense_Button(hri_window,LAB_Exit,TRUE);
    Sense_Button(hri_window,LAB_Save,TRUE);
    rdi->recording=FALSE;
    Set_Button(hri_window,LAB_Start,FALSE);
  }

  if ((!strcmp(name,LAB_Start)) && (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))))
  {
    rdi->recording=FALSE;
  }


  if (!strcmp(name,LAB_Exit))
  {
    Close_Window(hri_window);
    Sense_Button(main_window,MENU_HRPT,TRUE);
    Sense_Button(main_window,MENU_RAWD,TRUE);
  }
/* Write new select to decoder */
  if (reprogramming_fpga)
  {
/* Test if we have parallel port privilege */
    if (ioperm(addr_portd,3,1))
    {
      Create_Message("Error",PAR_ERR_MESSAGE,addr_portd);
      return;
    }

    if (!rdi->hwsel_type)
      rdi->sel_hw=rdi->gen1_dec0;

    select_channel_pp(NULL,rdi,FALSE);
    reprogramming_fpga=FALSE;
  }
}

static void close_wnd(GtkWidget *wnd)
{
  GtkWidget *mwnd=Find_Window(gtk_widget_get_toplevel(wnd),Lab_Main);;
  Sense_Button(mwnd,MENU_HRPT,TRUE);
  Sense_Button(mwnd,MENU_RAWD,TRUE);
  destroy_window(wnd);
}

/*************************************
 * Record window
 * This function is started from the main menu.
 *************************************/
void Menu_Rec_Hri(GtkWidget *widget)
{
  RECORDINFO *rdi=Get_Recordinfo_db(widget);
  GtkWidget *w1,*w2;
  GtkWidget *wa,*wb,*wc,*wd,*we,*wh,*wq,*wx,*wy,*wz;
  GtkWidget *hri_window;
  GtkWidget *main_window=Find_Parent_Window(widget);
  GdkColor startbut_clr,startbutsclr;
  startbut_clr.red=0xffff; startbut_clr.green=0x5555; startbut_clr.blue=0x5555;
  startbutsclr.red=0xffff; startbutsclr.green=0x8888; startbutsclr.blue=0x8888;

  if (!(hri_window=Create_Window(main_window,0,0,Lab_Hriwnd,close_wnd))) return;

/* Set policy on auto-resize */
  gtk_window_set_policy((GtkWindow *)hri_window,FALSE,FALSE,TRUE);

  hide_but=NULL;
  rdi->sat_type=HRI;

/* ==================== Button groups ==================== */
/* --------------------START/STOP----------------------- */

  w1=Create_Toggle(LAB_Start,Record_Hri_Commands,FALSE);
 Set_Color(w1,GTK_RC_BG,NULL,&startbutsclr,&startbut_clr,&startbut_clr,NULL);

/*
  w2=Create_Button(LAB_Stop,Record_Hri_Commands);
*/
  wa=w1;

/* --------------------- STATUS ------------------------- */
  w1=Create_Entry(LAB_Recording,NULL,"Off        ");
  w2=Create_Entry(LAB_Satname  ,NULL,"           ");
  wb=Pack(NULL,'v',w1,0,w2,0,NULL);

  w1=Create_Entry(LAB_RECLINES,NULL,"%4d",0);
  w2=Create_Entry(LAB_Encryp  ,NULL," ");
  wc=Pack(NULL,'h',w1,0,w2,0,NULL);

  w1=Create_Entry(LAB_Cntrec   ,NULL,"  ");
  w2=Create_Entry(LAB_Cntencr  ,NULL,"  ");
  wd=Pack("Received up to now",'h',w1,0,w2,0,NULL);

  wb=Pack("Status",'v',wb,5,wc,5,wd,5,NULL);


/* ==================== Advanced Button groups ==================== */
/* --------------------- DEC/GEN ---------------------- */
  if (rdi->hwsel_type)
  {
    wc=NULL;
  }
  else
  {
    w1=Create_Radio(LAB_DEC,TRUE,Record_Hri_Commands);
    w2=Create_Radio(LAB_GEN,FALSE,Record_Hri_Commands);
    wc=Pack("Hardware select",'h',w1,0,w2,0,NULL);
  }


/* --------------------SAVE----------------------- */
  w1=Create_Check(LAB_Save,Record_Hri_Commands,rdi->save_flag);
  w2=Create_Check(LAB_RECALL,Record_Hri_Commands,rdi->record_all);
  wd=Pack("",'h',w1,10,w2,10,NULL);

/* --------------------(Advanced) / Exit----------------------- */
  we=Create_Button(LAB_Exit,Record_Hri_Commands);
  wq=Create_Toggle(LAB_ADVANCED,Record_Hri_Commands,FALSE);
  wh=SPack("","Hh",wq,"fe5",we,"fe5",NULL);

  wx=Pack(NULL,'v',wa,5,wb,5,wh,5,NULL);
  wy=Pack(LAB_ADV_SET,'v',wc,5,wd,1,NULL);

  wz=Pack(NULL,'v',wx,5,wy,5,NULL);

  gtk_container_add(GTK_CONTAINER(hri_window),wz);

  gtk_widget_show(hri_window);
  place_window(hri_window,0,0,prefs.sub_wnd_geom);

  if (rdi->port_type==parport)
  {
    if (!ioperm(rdi->parport_base,3,1))
    {
      if (get_ppreg(rdi->parport_base+2) & PIN_17)
        rdi->sel_hw=FALSE;
      else
        rdi->sel_hw=TRUE;

      if (rdi->hwsel_type)
      {
        rdi->sel_hw=TRUE;
        select_channel_pp(NULL,rdi,FALSE);
      }
      else
        Set_Button(hri_window,LAB_GEN,rdi->sel_hw);
    }
  }


  Record_Hri_Commands(wq,LAB_ADVANCED);

  Sense_Button(main_window,MENU_HRPT,FALSE);
  Sense_Button(main_window,MENU_RAWD,FALSE);
}

/*************************************
 * Record auto. 
 *************************************/
void Recordpp_Hri_Auto(GtkWidget *main_window,gboolean generator)
{
  GtkWidget *recwnd;

  if (main_window) Menu_Rec_Hri(main_window);
  recwnd=Find_Window(main_window,Lab_Hriwnd);
  Set_Button(recwnd,LAB_DEC,1);

  if (generator) Set_Button(recwnd,LAB_GEN,1);

  Set_Button(recwnd,LAB_Start,1);
  Record_Hri_Commands(Find_Widget(recwnd,LAB_Start), LAB_Start);
}
#endif
