/**************************************************
 * RCSId: $Id$
 *
 * Old HRPT reader
 * Project: WSAT
 * Author: R. Alblas
 * NOTE: This function is not guaranteed portable between big/little endian
 *       machines!
 *
 * History: 
 * $Log$
 **************************************************/
/*******************************************************************
 * Copyright (C) 2000 R. Alblas. 
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software. If not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 ********************************************************************/

#include <math.h>
#include <malloc.h>
#include <time.h>

#include "wsat.h"
#include "hrpt.h"
#include "hrpt1.h"

/*************************************
 * Read HRPT1 format header
 * into PICINFO struct.
 * Return value:
 *   pointer to struct with header info
 *   NULL if failed
 *************************************/

gboolean check_tm(struct tm *);
/*************************************
 * Read hrpt/seawifs header into struct
 *************************************/
static HRPTHDR *read_hrpt1hdr_file(FILE *fp)
{
  HRPTHDR *hrpthdr;

  if (fp==NULL) return NULL;
  rewind(fp);

/* allocate mem for hrpthdr struct */
  if ((hrpthdr=(HRPTHDR *)calloc(sizeof(HRPTHDR),1))==NULL) return NULL;

/* Read hrpt header in 4 steps, to make it portable between DOS and Linux */

/* Read first part: sign ...  tm_comp_sat_vlg */
  if (!(fread(hrpthdr,131,1,fp)))
  { free(hrpthdr);  return NULL; }

/* Read second part: offset_lon ... bo */
  if (!(fread((char *)&hrpthdr->offset_lon,42,1,fp)))
  { free(hrpthdr);  return NULL; }

/* Read third part: kr[0] ... channel[9]*/
  if (!(fread((char *)&hrpthdr->kr[0],105,1,fp)))
  { free(hrpthdr);  return NULL; }

/* Read fourth part: widthmax ... end */
  if (!(fread((char *)&hrpthdr->widthmax,15722,1,fp)))
  { free(hrpthdr);  return NULL; }

/* Convert version 3 header to version 4 */
  if (*(hrpthdr->hdr_version+7) == '3')
  {
    int i;
    for (i=0; i<=8; i++) hrpthdr->channel[i]=FALSE;
    if (hrpthdr->kan_tot&0x02) hrpthdr->channel[0]=TRUE;
    if (hrpthdr->kan_tot&0x40) hrpthdr->channel[1]=TRUE;
    if (hrpthdr->kan_tot&0x20) hrpthdr->channel[2]=TRUE;
    if (hrpthdr->kan_tot&0x10) hrpthdr->channel[3]=TRUE;
    if (hrpthdr->kan_tot&0x08) hrpthdr->channel[4]=TRUE;
    if (hrpthdr->kan_tot&0x04) hrpthdr->channel[5]=TRUE;
  }

  return hrpthdr;
}

/*************************************
 * Read chrpt header into struct
 *************************************/
static CHRPTHDR *read_chrpt1hdr_file(FILE *fp)
{
  CHRPTHDR *chrpthdr;

  if (fp==NULL) return NULL;
  rewind(fp);

/* allocate mem for hrpthdr struct */
  if ((chrpthdr=(CHRPTHDR *)calloc(sizeof(CHRPTHDR),1))==NULL) return NULL;

/* Read chrpt header in 4 steps, to make it portable between DOS and Linux */

/* Read first part: sign ...  tm_comp_sat_vlg */
  if (!(fread(chrpthdr,131,1,fp)))
  { free(chrpthdr);  return NULL; }

/* Read second part: offset_lon ... bo */
  if (!(fread((char *)&chrpthdr->offset_lon,50,1,fp)))
  { free(chrpthdr);  return NULL; }

/* Read third part: kr[0] ... channel[11]*/
  if (!(fread((char *)&chrpthdr->kr[0],131,1,fp)))
  { free(chrpthdr);  return NULL; }

/* Read fourth part: widthmax ... end */
  if (!(fread((char *)&chrpthdr->widthmax,15688,1,fp)))
  { free(chrpthdr);  return NULL; }

  return chrpthdr;
}

/*************************************
 * Calculate # bytes per frame for HRPT-related formats
 *************************************/
FRAMEINFO calc_hrptframeinfo(FILE_TYPE file_type,gboolean *channels,int width)
{
  FRAMEINFO fri;
  int i;
  fri.header_len=0;
  fri.leader_len=0;

/* === Determine nr. of non-picture data === */
  if (channels[0])
  {
/* frame header present */
    switch (file_type)
    {
      case HRPT1:
        fri.header_len=750;
      break;
      case CHRPT1:
        fri.header_len=1600;
      break;
      case SEAWIF1:
        fri.header_len=685;
      break;
      default: break;
    }
    fri.leader_len=100;
  }
  else
  {
/* No frame header present */
    fri.header_len=0;
    fri.leader_len=0;
  }

/* === Determine words per line === */
  fri.words_per_frame=0;

/* Contribution of actual picture data */
  fri.inter_chan=0;
  for (i=1; i<NR_CHANNELS_MAX; i++)
    if (channels[i]) fri.inter_chan++;

  fri.words_per_frame=width*fri.inter_chan;

/* Contribution of header and leader. 
   Leader contains extra bytes to make words_per_frame*10/8 integer.
*/
  if (channels[0])
  {
    fri.words_per_frame+=fri.header_len;
    fri.words_per_frame+=fri.leader_len;
  }
  fri.words_per_line=fri.words_per_frame+((4-(fri.words_per_frame&0x3))&0x3);

/* Calculate file bytes per line */
  fri.bytes_per_line=fri.words_per_line*10/8;

/* Add contribution of extra 'frame separation' bytes */
  fri.bytes_per_line+=NR_INTER_BYTES;

  return fri;
}

/*************************************
 * Read a quad word (5 bytes)
 *************************************/
static int read_qwords(FILE *fp,guint16 *pline,int nr_qwords)
{
  int i,j;
  guchar tmp[7];
  for (i=0; i<nr_qwords; i++)
  {
    if (!(fread(tmp,5,1,fp))) return 0;
    if (pline)
    {
      for (j=0; j<4; j++)
      {
        *(pline)=(tmp[j]<<2) + ((tmp[4] >> ((3-j)*2) )&0x3);
        pline++;
      }
    }
  }
  return 1;
}


/*************************************
 * Calculate # of lines in hrpt file
 *************************************/
static int hrpt_lines_in_file(PICINFO *pci)
{
  int y=0;
  FILE *fp=pci->fp;
  int nr_qwords=pci->fri.words_per_frame/4;
  fseek(fp,pci->pic_strt,SEEK_SET);
  for (; ; y++)
  {
    if (!(read_qwords(fp,NULL,nr_qwords))) break;
    if (!(read_qwords(fp,NULL,2))) break;
  }
  return y;
}


/*************************************
 * Read (c)hrpt header and init PICINFO struct
 *************************************/
int read_hrpt1_hdr(PICINFO *pci)
{
  FILE *fp=pci->fp;
  HRPTHDR *hrpthdr;
  CHRPTHDR *chrpthdr;
  int i;

/* Read header for hrpt or chrpt */
  switch(pci->file_type)
  {
    case CHRPT1:
      if (!(chrpthdr=read_chrpt1hdr_file(fp))) return 'r';
      hrpthdr=(HRPTHDR *)chrpthdr;
    break;
    default:
      if (!(hrpthdr=read_hrpt1hdr_file(fp))) return 'r';
    break;
  }

/* Copy size info into picinfo structure */
  pci->width=GUINT16_FROM_LE(hrpthdr->width);
  pci->height=GUINT16_FROM_LE(hrpthdr->height);
  pci->depth=10;

/* Copy color map into picinfo structure */
  pci->clrmap.size=1024;

  for (i=0; i<pci->clrmap.size; i++)
  {
    pci->clrmap.red[i]=i;
    pci->clrmap.green[i]=i;
    pci->clrmap.blue[i]=i;
  }

/* Determine channels present */
  for (i=0; i<NR_CHANNELS_MAX; i++) pci->channels_recorded[i]=FALSE;

  pci->lines_per_sec=6;
  switch(pci->file_type)
  {
    case HRPT1:                /* HRPT */
      for (i=0; i<=5; i++)
        pci->channels_recorded[i]=hrpthdr->channel[i];
      pci->width_original=2048;
      pci->max_sens_angle=DEF_NOAA_ALPHAMAX;
      strcpy(pci->sat_name,hrpthdr->sat_name);
      pci->sat_type=HRPT;
    break;
    case CHRPT1:               /* CHRPT */
      for (i=0; i<=10; i++) 
        pci->channels_recorded[i]=chrpthdr->channel[i];
      pci->width_original=2048;
      pci->max_sens_angle=DEF_NOAA_ALPHAMAX;
      strcpy(pci->sat_name,chrpthdr->sat_name);
      pci->sat_type=CHRPT;
    break;
    case SEAWIF1:              /* Seawifs */
      for (i=0; i<=8; i++)
        pci->channels_recorded[i]=hrpthdr->channel[i];
      pci->width_original=1288;
      pci->max_sens_angle=DEF_FY_ALPHAMAX;
      strcpy(pci->sat_name,hrpthdr->sat_name);
      pci->sat_type=SEASTAR;
    break;
    default:
    break;
  }

/* Start of picture here */
  pci->pic_strt=ftell(fp)+7;
  pci->fri=calc_hrptframeinfo(pci->file_type,pci->channels_recorded,pci->width);

/* If no height -> calc. it (record was aborted?) */
  if (!pci->height)
  {
    pci->height=hrpt_lines_in_file(pci);
  }
  
/* Satellite scan direction */
  if (hrpthdr->satdir)
  {
    pci->scan_hor=E_W;
    pci->scan_ver=S_N;
  }
  else
  {
    pci->scan_hor=W_E;
    pci->scan_ver=N_S;
  }
  pci->ox=hrpthdr->x_offset;
  pci->oy=hrpthdr->y_offset;

/* Copy Keplerian elements */
  pci->kepler.epoch_year=GUINT32_FROM_LE(hrpthdr->kepler.epoch_year);
  pci->kepler.epoch_day=float32_from_le(hrpthdr->kepler.epoch_day);
  pci->kepler.decay_rate=float32_from_le(hrpthdr->kepler.decay_rate);
  pci->kepler.inclination=D2R(float32_from_le(hrpthdr->kepler.inclination));
  pci->kepler.raan=D2R(float32_from_le(hrpthdr->kepler.raan));
  pci->kepler.eccentricity=float32_from_le(hrpthdr->kepler.eccentricity);
  pci->kepler.perigee=D2R(float32_from_le(hrpthdr->kepler.perigee));
  pci->kepler.anomaly=D2R(float32_from_le(hrpthdr->kepler.anomaly));
  pci->kepler.motion=float32_from_le(hrpthdr->kepler.motion);
  pci->kepler.epoch_rev=GUINT32_FROM_LE(hrpthdr->kepler.epoch_rev);

/* starttime from satellite */
  pci->orbit.start_tm_sat.tm_year=GUINT16_FROM_LE(hrpthdr->tm_satstrt.tm_year);
  pci->orbit.start_tm_sat.tm_yday=GUINT16_FROM_LE(hrpthdr->tm_satstrt.tm_yday);
  pci->orbit.start_tm_sat.tm_hour=GUINT16_FROM_LE(hrpthdr->tm_satstrt.tm_hour);
  pci->orbit.start_tm_sat.tm_min=GUINT16_FROM_LE(hrpthdr->tm_satstrt.tm_min);
  pci->orbit.start_tm_sat.tm_sec=GUINT16_FROM_LE(hrpthdr->tm_satstrt.tm_sec);
  pci->orbit.start_tm_sat.tm_mon=GUINT16_FROM_LE(hrpthdr->tm_satstrt.tm_mon);
  pci->orbit.start_tm_sat.tm_mday=GUINT16_FROM_LE(hrpthdr->tm_satstrt.tm_mday);
  pci->orbit.start_tm_sat.tm_isdst=GUINT16_FROM_LE(hrpthdr->tm_satstrt.tm_isdst);

/* starttime from computer */
  pci->orbit.start_tm_comp.tm_year=GUINT16_FROM_LE(hrpthdr->tm_comp.tm_year);
  pci->orbit.start_tm_comp.tm_yday=GUINT16_FROM_LE(hrpthdr->tm_comp.tm_yday);
  pci->orbit.start_tm_comp.tm_hour=GUINT16_FROM_LE(hrpthdr->tm_comp.tm_hour);
  pci->orbit.start_tm_comp.tm_min=GUINT16_FROM_LE(hrpthdr->tm_comp.tm_min);
  pci->orbit.start_tm_comp.tm_sec=GUINT16_FROM_LE(hrpthdr->tm_comp.tm_sec);
  pci->orbit.start_tm_comp.tm_mon=GUINT16_FROM_LE(hrpthdr->tm_comp.tm_mon);
  pci->orbit.start_tm_comp.tm_mday=GUINT16_FROM_LE(hrpthdr->tm_comp.tm_mday);
  pci->orbit.start_tm_comp.tm_isdst=GUINT16_FROM_LE(hrpthdr->tm_comp.tm_isdst);

/* Check times in header. These should contain valid values! 
   Invalid times will be forced to valid but wrong ones. 
   (mktime can take very long time if values are far out of range!)
*/
  check_tm(&pci->orbit.start_tm_sat);
  check_tm(&pci->orbit.start_tm_comp);

  pci->orbit.offset_gmt=hrpthdr->offset_gmt;
  pci->orbit.offset_sec=hrpthdr->offset_sec;
  pci->orbit.offset_lon=hrpthdr->offset_lon;
  pci->orbit.use_sattime=(hrpthdr->tm_comp_sat_vlg=='s'? TRUE : FALSE);
  free(hrpthdr);

/* Copy starttime to use */
  if (pci->orbit.use_sattime)
  {
    pci->orbit.start_tm=pci->orbit.start_tm_sat;
    pci->orbit.start_ms=pci->orbit.start_ms_sat;
  }
  else
  {
    pci->orbit.start_tm=pci->orbit.start_tm_comp;
    pci->orbit.start_ms=0;
  }

/* Calculate some orbit constants */
  calc_orbitconst(&pci->kepler,&pci->orbit);
  return 0;
}


/*************************************
 * Put (c)hrpt line number y from file into *line
 * Return: decoder frame error code
 *************************************/
int read_hrpt1_line(PICINFO *pci,VIEWINFO *vwi,int y,guint16 *line)
{
  FILE *fp=pci->fp;
  guchar tmp[7];
  guint16 syncpost[8];
  int nr_qwords=pci->fri.words_per_frame/4;
  int rslt;

  do
  {
    rslt=MASK_HRPT1_READ_OK_PRE;

    fseek(fp,pci->pic_strt + y*pci->fri.bytes_per_line,SEEK_SET);

    read_qwords(fp,line,nr_qwords);

/* Read extra bytes for decoder info 
   sync       in-sync -> going_out -> out-sync -> going_in -> sync
   dec-status 1          1            0           0           1
y+1 -> take sync-info after current line
y+0 -> take sync-info just before current line
y-1 -> take sync-info before one line prior to current line
5-3-2004: syncsignal_del changed from +1 to -1
*/

    fseek(fp,pci->pic_strt + (y+1)*pci->fri.bytes_per_line-NR_INTER_BYTES,
             SEEK_SET);

    if (!(fread(tmp,NR_INTER_BYTES,1,fp))) return 0;

/* read next sync */
    read_qwords(fp,syncpost,2);

/* Test sync mode if sync-word is recorded */
    if (pci->channels_recorded[0])
    {
      if ((line[0]==0x284) && (line[1]==0x16f) && (line[2]==0x35c) &&
          (line[3]==0x19d) && (line[4]==0x20f) && (line[5]==0x095))
        rslt|=MASK_HRPT1_SYNC_OK_PRE;

      if ((syncpost[0]==0x284) && (syncpost[1]==0x16f) && (syncpost[2]==0x35c) &&
          (syncpost[3]==0x19d) && (syncpost[4]==0x20f) && (syncpost[5]==0x095))
        rslt|=MASK_HRPT1_SYNC_OK_POST;
    }
    else  /* use decoder status if no frame header is recorded */
    {
      if (tmp[4]=='E') rslt|=MASK_HRPT1_SYNC_OK_PRE; /* No during record */
    }

    y=y+(vwi->scan_ver==S_N? (-1) : (1));   /* incr or decr line nr. dep. on scan dir. */
    if (y<0) return 0;                      /* stop if BOF */

    if (!vwi->skip_faulty_lines) break;
    if ( ((rslt&MASK_HRPT1_SYNC_OK_PRE)  || (!(vwi->mask_skip&MASK_HRPT1_SYNC_OK_PRE))) &&
         ((rslt&MASK_HRPT1_SYNC_OK_POST) || (!(vwi->mask_skip&MASK_HRPT1_SYNC_OK_POST))) )
      break;

  } while (rslt);                           /* stop if EOF */

  return rslt;
}
