/**************************************************
 * RCSId: $Id$
 *
 * Data processing functions.
 * Project: WSAT
 * Author: R. Alblas
 *
 * History: 
 * $Log$
 **************************************************/
/*******************************************************************
 * Copyright (C) 2000 R. Alblas. 
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software. If not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 ********************************************************************/

#include <malloc.h>
#include "wsat.h"

#if _SUN_
#  include "sun.h"
#endif
#if _TIFF_
#  include "tiff.h"
#endif
#if _PPM_
#  include "ppm.h"
#endif
#if _GIF_
#  include "gif.h"
#endif
#if _HRPT1_
#  include "hrpt1.h"
#endif
#if _HRI1_
#  include "hri1.h"
#endif

/*************************************
 * Get line 'y' from file in any format
 * NOTE: HRI1 : channel is selected here.
 *       HRPT1: all channels are read (selection lateron).
 *************************************/
int read_line(PICINFO *pci,VIEWINFO *vwi,int y,
              guint16 *l,guint16 *l2,guint16 *l3,gboolean *line_grid)
{
  int dec_status=1;
  guchar grid[GRIDBYTES_LINE];     /* Each byte has gridinfo for 8 pixels */

  switch(pci->file_type)
  {
#if _SUN_
    case SUN:
      read_sun_line(pci,y,l);
    break;
#endif
#if _TIFF_
    case TIFF:
      read_tiff_line(pci,y,l,l2,l3);
    break;
#endif
#if _PPM_
    case PPM: case PGM:
      if (!(read_ppm_line(pci,y,l,l2,l3)));
    break;
#endif
#if _GIF_
    case GIF:
      read_gif_line(pci,y,l);
    break;
#endif
#if _HRPT1_
    case HRPT1: case CHRPT1: case SEAWIF1:
      dec_status=read_hrpt1_line(pci,vwi,y,l);
    break;
#endif
#if _HRI1_
    case HRI1:
      dec_status=read_hri1_line(pci,vwi,y,l,grid);

/* If grid requested then add gridinfo to 'line_grid'; 
   this will act as a mask to add gridinfo into picture
*/

      if (vwi->grid) add_hrigrid(pci,line_grid,grid);
    break;
#endif
#if _AHRPT1_
    case AHRPT1:
      dec_status=read_ahrpt1_line(pci,vwi,y,l);
    break;
#endif
    default: break;
  }
  return dec_status;
}

void despeckle(PICINFO *pci,VIEWINFO *vwi,guint16 **lines,int nr_lines)
{
  guint16 *pix1,*pix2,*pix3;
  guint16 du,dd,dl,dr;
  int inter_chan=pci->fri.inter_chan;
  int nr_1lines=nr_lines/3;
  int nrl,c,x;
  int sur_ulim=vwi->sur_ulim;
  int dsp_llim=vwi->dsp_llim;
  int l_offset;

  for (nrl=0; nrl<nr_1lines; nrl++)
  {
    for (x=1; x<pci->width-1; x++)
    {
      int ci=0;
      for (c=1; c<NR_CHANNELS_MAX; c++)         /* Do despeckling for all channels */
      {
        if (pci->channels_recorded[c])
        {
          l_offset=pci->fri.header_len+ci+x*inter_chan;
          pix1=lines[nrl]+l_offset;
          pix2=lines[nrl+nr_1lines]+l_offset;
          pix3=lines[nrl+2*nr_1lines]+l_offset;
          du=ABS(*pix2-*pix1);
          dd=ABS(*pix2-*pix3);
          dl=ABS(*pix2-*(pix2-inter_chan));
          dr=ABS(*pix2-*(pix2+inter_chan));
          if ((ABS(du-dd) < sur_ulim) && 
              (ABS(dl-dr) < sur_ulim) && 
              (ABS(du-dl) < sur_ulim) &&
              (du > dsp_llim))
          {
            *pix1=*pix1;
          }
          else
          {
            *pix1=*pix2;
          }
          ci++;
        } /* if */
      } /* for */
    }
  }
}
 
/**********************************************************
 * Read 1 or more lines
 **********************************************************/
int read_lines(VIEWINFO *vwi,PICINFO *pci,
                guint *ymap,int y,int nr_lines,
                guint16 **lines,guint16 *line_g,guint16 *line_b,
                gboolean *line_grid)
{
  int i;
  int dec_status=0;
  static int ymp=-1;
  int ym;

/* If despcikle: Read lines n-1, n and n+1 (nr_ilines =3x normal) */


/* Determine already readed lines and lines to read */
  ym =*(ymap+y);

/* Next used to be shift to compensate shift added by despeckling.
   Removed because grid-info has also to be shifted.
   This has to be solved in a better way!
*/

/*
  if (vwi->despeckle)
  {
    ym=MAX(ym-1,0);
  }
*/


/* If same, nothing to read. */
  if (ym==ymp) return dec_status;

/* Shift already readed lines to correct place. 
   E.g. phase 1: readed lines 11, 12 and 13
        phase 2: needed 12, 13 and 14.
        ==> 12,13 -> 11,12 and only line 14 is read.

   This is needed for de-speckling GIF pictures because with current
   GIF functions it is not possible to read e.g. line 12 if line 13 is
   already read, without rewinding and re-reading all lines 1 ... 12 again.
   (Which is, of course, very very slow.)
   For other formats this method doesn't speed up reading because the OS
   (at least, Unix) is already caching readed lines.
   (This preventing of re-reading lines only works for north-south because
   it makes only sense for GIF, and GIF reading is currently only possible
   from begin to end, not vice versa)
*/

  if ((y>0) && (nr_lines>1) && (ym>ymp))
  {
    guint16 *tmp;
    for (i=0; i<nr_lines-(ym-ymp); i++)
    {
      tmp=lines[i];
      lines[i]=lines[i+ym-ymp];
      lines[i+ym-ymp]=tmp;
    }
  }
  else
  {
    i=0;                                   /* read all lines */
  }

  for (; i<nr_lines; i++)                  /* Now read all remaining lines */
  {
    dec_status|=read_line(pci,vwi,ym+i,lines[i],line_g,line_b,line_grid);
  }

  if (vwi->despeckle) despeckle(pci,vwi,lines,nr_lines);

  ymp=ym;
  return dec_status;
}

guint16 **alloc_lines(int size_line,int nr_lines)
{
  int i;
  guint16 **line_arr;
  line_arr=(guint16 **)calloc(sizeof(*line_arr),nr_lines);
  if (!line_arr) return NULL;
  for (i=0; i<nr_lines; i++)
  {
    line_arr[i]=(guint16 *)malloc(size_line*sizeof(*line_arr));
    if (!line_arr[i]) return NULL;
  }
  return line_arr;
}

void free_lines(guint16 **line_arr,int nr_lines)
{
  int i;
  if (!line_arr) return;
  for (i=0; i<nr_lines; i++)
  {
    if (line_arr[i]) free(line_arr[i]);
  }
  free(line_arr);
}


/**********************************************************
 * Translate line in b/w from file to line in r/g/b to draw.
 * 
 **********************************************************/

/**********************************************************
 * Translate pixel 'pix' into rgb using linear combination of channels.
 * 
 **********************************************************/
void Translate_Pix(int x,VIEWINFO *vwi,
                   PICINFO  *pci,
                   guint16  **lines,
                   int      nrlines,
                   int      pix,
                   guint16  *r,
                   guint16  *g,
                   guint16  *b)
{
  int c,c1=0;
  guint16 *pix1;
  int r1,g1,b1;
  int inter_chan=pci->fri.inter_chan;
  int vshift;

/* Use signed integers so neg. values after calc. can be clipped to 0. */
  r1=vwi->red_offset*4;
  g1=vwi->green_offset*4;
  b1=vwi->blue_offset*4;

  for (c=1; c<NR_CHANNELS_MAX; c++)
  {
    if (pci->channels_recorded[c])
    {
      if (vwi->channels_selected[c])
      {
        vshift=((vwi->tilt[c]*2*x)/pci->width-vwi->tilt[c])+
                                  vwi->vshift[c]-vwi->min_vshift;

/* abort if vshift points to not-readed line */
        if (vshift<0) break;
        if (vshift>=nrlines) break;

        pix1=lines[vshift]+pix+(vwi->hshift[c]*inter_chan)+c1;
/* Add contribution pix to rgb */
        (r1)+=(*pix1)*vwi->red[c];
        (g1)+=(*pix1)*vwi->green[c];
        (b1)+=(*pix1)*vwi->blue[c];
      }
      c1++;
    }
  }
  r1=MIN(r1,1023); r1=MAX(r1,0);
  g1=MIN(g1,1023); g1=MAX(g1,0);
  b1=MIN(b1,1023); b1=MAX(b1,0);
  *r=r1; *g=g1; *b=b1;
}

void Translate_Line(VIEWINFO  *vwi,         /* view info */
                    PICINFO   *pci,         /* picture info */
                    int       bitshift,
                    guint     *zmap,        /* color mapping */
                    gboolean  disable_xmir, /* disable mirror x-dir. */
                    guint16   **lines,      /* Input line */
                    int       nrlines,      /* nr. Input line */
                    guint16   *line_r,      /* Output red line */
                    guint16   *line_g,      /* Output green line */
                    guint16   *line_b)      /* Output blue line */
{
  int x1,x;
  guint16 r,g,b;
  int inter_chan=pci->fri.inter_chan;
  int header_len=pci->fri.header_len;

  for (x1=0; x1<pci->width; x1++)
  {
    if ((!disable_xmir) && (vwi->scan_hor==E_W))
    {
      x=pci->width-x1-1;
    }
    else
      x=x1;

    Translate_Pix(x1,vwi,pci,lines,nrlines,header_len+x1*inter_chan,&r,&g,&b);

/* Translate pix between 10bits <--> 8bits words */
    if (bitshift<0)
    {
      r>>=(bitshift*-1);
      g>>=(bitshift*-1);
      b>>=(bitshift*-1);
    }
    else
    {
       r<<=bitshift;
       g<<=bitshift;
       b<<=bitshift;
    }
    line_r[x]=r;
    line_g[x]=g;
    line_b[x]=b;
  }
}



/**********************************************************
 * Add grid. 
 * If vwi!=NULL -> take special grid colors.
 * If vwi =NULL -> use max. white ("0xff")
 **********************************************************/
void add_gridpix(VIEWINFO *vwi,guint16 *r,guint16 *g,guint16 *b)
{
  if (vwi)
  {
    *r=vwi->grid_clr.red;
    *g=vwi->grid_clr.green;
    *b=vwi->grid_clr.blue;
  }
  else
  {
    *r=0xff;
    *g=0xff;
    *b=0xff;
  }
}

