/**************************************************
 * RCSId: $Id$
 *
 * Draw related functions.
 * Project: WSAT
 * Author: R. Alblas
 *
 * History: 
 * $Log$
 **************************************************/
/*******************************************************************
 * Copyright (C) 2000 R. Alblas. 
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software. If not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 ********************************************************************/

#include <stdio.h>
#include <malloc.h>

#include "wsat.h"
#include "hrpt.h"

#if _SUN_
#  include "sun.h"
#endif
#if _TIFF_
#  include "tiff.h"
#endif
#if _PPM_
#  include "ppm.h"
#endif
#if _GIF_
#  include "gif.h"
#endif
#if _HRPT1_
#  include "hrpt1.h"
#endif
#if _HRI1_
#  include "hri1.h"
#endif


/**********************************************************
 * Draw line into pixmap.
 * Using gc/pixmap methodology.
 * Pixmap has to be drawn to canvas to make result visible. 
 **********************************************************/
void draw_a_line_gc(guint16    *line,        /* 1 line to draw */
                    int         wnd_width,   /* # pixels in line on canvas */
                    int         y,           /* y-position */
                    guint       *xmap,       /* hor. mapping */
                    guint       *zmap,       /* color mapping */
                    int         cmax,        /* Upper limit colormap */
                    GdkPixmap   *pixmap,     /* for double buffering */
                    GdkGC       **gc,        /* The set of gc's (color map) */
                    long        *lumstat)    /* array with lum. stat. to fill */
{
  int x,c;
  if (!pixmap) return;

/* For all hor. pixels in canvas */
  for (x=0; x<wnd_width; x++)
  {
    c=*(line+*(xmap+x));              /* Get byte */
    c=MIN(c,cmax);                    /* Clip lum */
    if (lumstat) lumstat[c]++;        /* Update luminance statistics */
    c=zmap[c];                        /* Translate file value to screen value */

    gdk_draw_point(pixmap,gc[c],x,y); /* Copy one pixel to buffer */
  }
}

/**********************************************************
 * Draw grid into rgb buffer
 * Using RGB methodology.
 * rgb buffer has to be drawn to canvas to make result visible. 
 **********************************************************/
void draw_grid(gboolean   *line_grid,   /* 1 grid-line to draw */
               int         wnd_width,   /* # pixels in line on canvas */
               int         y,           /* y-position */
               guint       *xmap,       /* hor. mapping */
               RGBI        *rgbi,       /* RGB picture buffer */
               VIEWINFO    *vwi)        /* view info */

{
  int x;
  guint16 r,g,b;
  int yoffset=y*wnd_width*3;

/* For all hor. pixels in canvas */
  for (x=0; x<wnd_width; x++)
  {
    if((*(line_grid+*(xmap+x)))) /* Extract grid info */
    {
      add_gridpix(vwi,&r,&g,&b);
      rgbi->rgbbuf[yoffset+3*x+0]=r;
      rgbi->rgbbuf[yoffset+3*x+1]=g;
      rgbi->rgbbuf[yoffset+3*x+2]=b;
    }
  }
}


/**********************************************************
 * Draw line into rgb buffer (black/white)
 * Using RGB methodology.
 * rgb buffer has to be drawn to canvas to make result visible. 
 * NOTE This is faster than gc/pixmap methodology!
 **********************************************************/
void draw_a_line_bw(guint16    *line,        /* 1 line to draw */
                    int         wnd_width,   /* # pixels in line on canvas */
                    int         y,           /* y-position */
                    guint       *xmap,       /* hor. mapping */
                    guint       *zmap,       /* color mapping */
                    int         cmax,        /* Upper limit colormap */
                    RGBI        *rgbi,       /* RGB picture buffer */
                    VIEWINFO    *vwi)        /* view info */
{
  guint x,c;
  int yoffset=y*wnd_width*3;
  long *lumstat=NULL;
  guint16 r,g,b;

  if (!rgbi->rgbbuf) return;

/* Abort if out-of-drawable */
  if (yoffset+3*(wnd_width-1)+2 >= rgbi->bufsize) return;

  if (vwi) lumstat=vwi->lumstat;

/* For all hor. pixels in canvas */
  for (x=0; x<wnd_width; x++)
  {
    c=*(line+*(xmap+x));              /* Get byte */
    c=MIN(c,cmax);                    /* Clip lum (also for lumstat!) */
    if (lumstat) lumstat[c]++;        /* Update luminance statistics */
    c=zmap[c];                        /* Translate file value to screen value */

    r=(rgbi->rgb_cmap->colors[c]>>16)&0xff;
    g=(rgbi->rgb_cmap->colors[c]>> 8)&0xff;
    b=(rgbi->rgb_cmap->colors[c]    )&0xff;

/* Copy one pixel to buffer */
    rgbi->rgbbuf[yoffset+3*x+0]=r;
    rgbi->rgbbuf[yoffset+3*x+1]=g;
    rgbi->rgbbuf[yoffset+3*x+2]=b;
  }
}

/**********************************************************
 * Draw line into rgb buffer (full color)
 * Using RGB methodology.
 * rgb buffer has to be drawn to canvas to make result visible. 
 **********************************************************/
void draw_a_line_rgb(guint16    *line_r,        /* 1 line to draw in red */
                     guint16    *line_g,        /* 1 line to draw in green */
                     guint16    *line_b,        /* 1 line to draw in blue */
                     int         wnd_width,     /* # pixels in line on canvas */
                     int         y,             /* y-position */
                     guint       *xmap,         /* hor. mapping */
                     guint       *zmap,         /* hor. mapping */
                     RGBI        *rgbi,         /* RGB picture buffer */
                     VIEWINFO    *vwi)          /* view info */
{
  int x;
  guint16 r,g,b;
  int yoffset=y*wnd_width*3;
  long *lumstat=NULL;

  if (!rgbi->rgbbuf) return;
  if (vwi) lumstat=vwi->lumstat;

/* For all hor. pixels in canvas */
  for (x=0; x<wnd_width; x++)
  {
    r=*(line_r+*(xmap+x));              /* Get red byte */
    g=*(line_g+*(xmap+x));              /* Get green byte */
    b=*(line_b+*(xmap+x));              /* Get blue byte */

    if (lumstat)
    {
      guint lm=(r+g+b)/3; lm=MIN(lm,1023);
      lumstat[lm]++;  
    }

/* Clip rgb */
    r=MIN(r,0x3ff);
    g=MIN(g,0x3ff);
    b=MIN(b,0x3ff);

    if (zmap)
    {
      r=zmap[r];
      g=zmap[g];
      b=zmap[b];
    }

    r=MIN(r,255);
    g=MIN(g,255);
    b=MIN(b,255);

    r=(rgbi->rgb_cmap->colors[r]>>16)&0xff;
    g=(rgbi->rgb_cmap->colors[g]>> 8)&0xff;
    b=(rgbi->rgb_cmap->colors[b]    )&0xff;

/* Copy one pixel to buffer */
    rgbi->rgbbuf[yoffset+3*x+0]=r;
    rgbi->rgbbuf[yoffset+3*x+1]=g;
    rgbi->rgbbuf[yoffset+3*x+2]=b;
  }
}

/**********************************************************
 * Draw line into rgb buffer.
 * Use function 'map_function' to convert pixels into rgb.
 * Input map_function : pixel-value and lum. map
 * Output map_function: r/g/b
 *
 * Using RGB methodology.
 * rgb buffer has to be drawn to canvas to make result visible. 
 **********************************************************/
void draw_a_line_func(guint16    *line,        /* 1 line to draw */
                      GtkWidget *drawing_area,
                      int         wnd_width,   /* # pixels in line on canvas */
                      int         y,           /* y-position */
                      guint       *xmap,       /* hor. mapping */
                      guint       *zmap,       /* color mapping */
                      int         cmax,        /* Upper limit colormap */
                      RGBI        *rgbi,       /* RGB picture buffer */
                      VIEWINFO    *vwi,        /* view info */
                      gint        map_function())
{
  gint x,c;
  int yoffset=y*wnd_width*3;
  long *lumstat=NULL;
  guint16 r,g,b;
  if (!rgbi->rgbbuf) return;

  if (vwi) lumstat=vwi->lumstat;

/* For all hor. pixels in canvas */
  for (x=0; x<wnd_width; x++)
  {
    c=*(line+*(xmap+x));              /* Get byte */

    c=map_function(c,zmap,&r,&g,&b);
    if (lumstat) lumstat[c]++;        /* Update luminance statistics */

    r=(rgbi->rgb_cmap->colors[r]>>16)&0xff;
    g=(rgbi->rgb_cmap->colors[g]>> 8)&0xff;
    b=(rgbi->rgb_cmap->colors[b]    )&0xff;


/* Copy one pixel to buffer */
    rgbi->rgbbuf[yoffset+3*x+0]=r;
    rgbi->rgbbuf[yoffset+3*x+1]=g;
    rgbi->rgbbuf[yoffset+3*x+2]=b;
  }
}

/*************************************
 * Draw the picture into the canvas
 * Return TRUE is draw_pic was aborted.
 *************************************/
gboolean draw_pic(PICINFO   *pci,
                  VIEWINFO  *vwi,
                  GtkWidget *window)
{
  GtkWidget *drawing_area=Find_Widget(window,"GTK_DRAWING_AREA");
  RGBI *rgbi=Get_RGBI(drawing_area);

  int y,yp,i;
  guint c;
  int chan_sel=0;
  int dec_status=0;
  int wnd_width =drawing_area->allocation.width;
  int wnd_height=drawing_area->allocation.height;
  int nrclr;
  int nr_ilines;           /* nr. lines needed to produce 1 output line */
  guint16 *line;
  guint16 **lines;
  guint16 *line_r=NULL,*line_g=NULL,*line_b=NULL;
  gboolean *line_grid=NULL;
  guint *xmap=NULL;
  guint *ymap=NULL;
  guint *zmap=NULL;
  gboolean map_tempflag=FALSE;
  NOAACAL noaacal;

  nrclr=pci->clrmap.size;
  if ((IS_HRPTIR(vwi)) && (vwi->map_temp)) map_tempflag=TRUE;

  if (pci->fp==NULL) return FALSE; 

/* Do processing for some types of file formats */
#if _GIF_
  if (pci->file_type==GIF) rewind_gif(pci);
#endif
#if _HRI1_
  if (pci->file_type==HRI1) hri_channelselect(pci,vwi);
#endif

/* Check sizes */
  if (!pci->width) return FALSE;
  if (!pci->height) return FALSE;

/* Calculate mapping. Contains also channel select for multi-chan. files. */
  if (map_pic(window,pci,vwi,vwi->process_data,&xmap,&ymap,&zmap)) return TRUE;

/* Determine which channel is selected (--> chan_sel) */
  for (i=1; i<NR_CHANNELS_MAX; i++)
  {
    if ((vwi->channel_selected[i]) && (pci->channels_recorded[i])) 
      break;
    if (pci->channels_recorded[i]) chan_sel++;
  }


/* determine how many lines are needed for processing */
  vwi->min_vshift=vwi->max_vshift=0;
  if (vwi->process_data)
  {
    for (i=1; i<NR_CHANNELS_MAX; i++)
    {
      vwi->min_vshift=MIN(vwi->min_vshift,(vwi->vshift[i]-ABS(vwi->tilt[i])));
      vwi->max_vshift=MAX(vwi->max_vshift,(vwi->vshift[i]+ABS(vwi->tilt[i])));
    }
  }


/* Prepare for drawing */
  if (map_tempflag) noaacal=load_cal(pci->sat_name);

/* nr. lines needed to produce 1 output line */
  nr_ilines=vwi->max_vshift-vwi->min_vshift+1;
  if (vwi->despeckle) nr_ilines*=3;
  if (map_tempflag) nr_ilines=MAX(nr_ilines,5);

  lines=(guint16 **)alloc_lines(pci->fri.words_per_line,nr_ilines);

  for (c=0; c<MAX_PCICLR; c++) vwi->lumstat[c]=0;

  line_r=(guint16 *)malloc(pci->fri.words_per_line*sizeof(*line_r));
  line_g=(guint16 *)malloc(pci->fri.words_per_line*sizeof(*line_g));
  line_b=(guint16 *)malloc(pci->fri.words_per_line*sizeof(*line_b));

  if ((pci->file_type==HRI1) && (vwi->grid))
    line_grid=(gboolean *)malloc(pci->fri.words_per_line*sizeof(*line_grid));


/* Draw the picture */
  Create_Cursor(drawing_area,GDK_WATCH);

  for (y=0; y<wnd_height; y++)          /* for all lines in current canvas */
  {
    if ((KeyPressed()) == GDK_Escape) break;  /* stop if escape pressed */
    while (g_main_iteration(FALSE));    /* Handle GTK actions now and then */

    if (vwi->redraw) break; /* Redraw requested, so abort current drawing */

/* Determine if a new line has to be read from file:
   . first line, or
   . line different from previous line (if zoomed in 1 line is repeated)
*/

    if ((y==0) || (*(ymap+y)!=*(ymap+yp)))
    {
      dec_status=read_lines(vwi,pci,ymap,y,nr_ilines,lines,line_g,line_b,line_grid);
      line=lines[0];
      yp=y;
    }

/* Choose between "process data" (calculate 1 result line from >1 source lines)
              and "fast" (normal, straight "read 1 line" and "plot") */
    if (vwi->process_data)
    {
/* Translate input line to rgb */
      Translate_Line(vwi,pci,0,zmap,TRUE,lines,nr_ilines,line_r,line_g,line_b);

/* Draw the line */
      draw_a_line_rgb(line_r,line_g,line_b,wnd_width,y,xmap,zmap,rgbi,vwi);
    }
    else if ((pci->file_type==PPM) || (pci->file_type==TIFF))
    {
/* Draw in color from TIFF or PPM format */
      draw_a_line_rgb(line,line_g,line_b,wnd_width,y,xmap,zmap,rgbi,vwi);
    }
    else if (map_tempflag)
    {
/* Draw IR temperature */
      if (yp==y)                 /* This happens if a new line is read */
        map_noaatemp(lines,&noaacal,pci,vwi);

      draw_a_line_func(line,drawing_area,wnd_width,y,
                       xmap,zmap,nrclr-1,rgbi,vwi,map_temp);
    }
    else
    {
/* Draw in B/W from any other format */
      draw_a_line_bw(line,wnd_width,y,xmap,zmap,nrclr-1,rgbi,vwi);
    }

/* Show faulty lines as a red bar */
#define MASK_PRESYNC  0x01
#define MASK_POSTSYNC 0x02
    if (vwi->mark_faulty_lines)
    {
      int oof=0;
      if (pci->file_type==HRI1)
      {
        if (!(dec_status & MASK_HRI1_SYNC_OK_POST)) oof|=MASK_PRESYNC;
      }
      if (IS_HRPT_FAM(pci->file_type))
      {
        if (!(dec_status & MASK_HRPT1_SYNC_OK_PRE))  oof|=MASK_PRESYNC;  /* 01 */
        if (!(dec_status & MASK_HRPT1_SYNC_OK_POST)) oof|=MASK_POSTSYNC; /* 10 */
      }

      if (oof)
      {
        int yoffset=y*wnd_width*3;
        int x;

        for (x=0; x<5; x++)
        {
          rgbi->rgbbuf[yoffset+3*x+0]=(oof&(MASK_PRESYNC|MASK_POSTSYNC)? 0xff : 0x00);
          rgbi->rgbbuf[yoffset+3*x+1]=(oof&MASK_POSTSYNC? 0xff : 0x00);
          rgbi->rgbbuf[yoffset+3*x+2]=0x00;
        }
      }
    }

/* Add grid; line_grid !=0 if vwi->grid = true */
    if (line_grid) draw_grid(line_grid,wnd_width,y,xmap,rgbi,vwi);


/* Make line just drawn visible */
    if (!vwi->refresh_pic)
      Refresh_Rect(drawing_area,0,y,wnd_width,1);

/* Start again if size window changed during plotting */
    if ((wnd_width!=drawing_area->allocation.width) ||
        (wnd_height!=drawing_area->allocation.height))
    {
      draw_pic(pci,vwi,window);
      break;
    }
  }

/* make picture visible */
  if (vwi->refresh_pic)
    Refresh_Rect(drawing_area,0,0,wnd_width,wnd_height);

  Create_Cursor(drawing_area,GDK_TOP_LEFT_ARROW);

  free_lines(lines,nr_ilines);
  if (line_r) free(line_r);
  if (line_g) free(line_g);
  if (line_b) free(line_b);
  if (line_grid) free(line_grid);
  free(xmap);
  free(ymap);
  free(zmap);
  return vwi->redraw;
}

