/**************************************************
 * RCSId: $Id$
 *
 * File related functions 
 * Project: WSAT
 * Author: R. Alblas
 *
 * History: 
 * $Log$
 **************************************************/
/*******************************************************************
 * Copyright (C) 2000 R. Alblas. 
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software. If not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 ********************************************************************/

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "wsat.h"
#include "menu_wsat.h"
#include "orbit.h"

#if _SUN_
#  include "sun.h"
#endif
#if _PPM_
#  include "ppm.h"
#endif
#if _TIFF_
#  include "tiff.h"
#endif
#if _GIF_
#  include "gif.h"
#endif
#if _HRPT1_
#  include "hrpt1.h"
#endif
#if _HRI1_
#  include "hri1.h"
#endif
#if _AHRPT1_
#  include "ahrpt.h"
#endif

#define LENSIGN 8       /* max expected length of signature of any format */

extern PREFS prefs;

/*************************************
 * Detect file format.
 * Return value:
 *   file_type
 *************************************/
FILE_TYPE detect_file(char *fn)
{

  char sign[LENSIGN];
#if _SUN_ || _TIFF_  || _AHRPT1_
  guint32 *sign1=(guint32 *)sign;
#endif

  FILE_TYPE file_type=unknown_filetype;
  FILE *fp;
  if ((fp=fopen(fn,"rb")) == NULL) return unknown_filetype;
  fread(sign,4,1,fp);

#if _SUN_
  if (GUINT32_FROM_BE(*sign1) == SUNSIGN)  file_type=SUN;
#endif
#if _TIFF_
  if (GUINT32_FROM_BE(*sign1) == TIFFSIGN) file_type=TIFF;
#endif
#if _PPM_
  if (!strncmp(sign,PPM_SIGN,strlen(PPM_SIGN))) file_type=PPM;
  if (!strncmp(sign,PGM_SIGN,strlen(PGM_SIGN))) file_type=PGM;
#endif
#if _GIF_
  if (!strncmp(sign,GIFSIGN,strlen(GIFSIGN))) file_type=GIF;
#endif
#if _HRPT1_
  if (!strncmp(sign,HRPT1SIGN,strlen(HRPT1SIGN))) file_type=HRPT1;
  if (!strncmp(sign,CHRPT1SIGN,strlen(CHRPT1SIGN))) file_type=CHRPT1;
  if (!strncmp(sign,SEAW1SIGN,strlen(SEAW1SIGN))) file_type=SEAWIF1;
#endif
#if _HRI1_
  if (!strncmp(sign,HRI1SIGN,4)) file_type=HRI1;
#endif
#if _AHRPT1_
  if (GUINT32_FROM_BE(*sign1) == AHRPT1SIGN)  file_type=AHRPT1;
#endif
  fclose(fp);

  return file_type;
}

char *strip_filename(char *fn)
{
  char *p;
  if ((p=strrchr(fn,'/'))) return p+1;
  if ((p=strrchr(fn,'\\'))) return p+1;
  if ((p=strrchr(fn,':'))) return p+1;
  return fn;
}

/*************************************
 * Open a file and detect format.
 * Read header info.
 * Return value:
 *   0   if OK
 *   !=0 if error
 *************************************/
int open_file(PICINFO *pci,
              VIEWINFO *vwi,
              GtkWidget *window)
{
  int err=0;
  FILE_TYPE file_type=unknown_filetype;

/* Detect file type. Exit if filetype unknown. */
  file_type=detect_file(pci->file_name);
  if (!file_type)
    return 'o';
  if (file_type==unknown_filetype)
    return 't';

  pci->file_type=file_type;
  pci->width_original=0;
/* Close previous file if needed, and open new file. */
  if (pci->fp) close_picfile(pci);
  if ((pci->fp=fopen(pci->file_name,"rb")) == NULL)
    return 'o';

  fseek(pci->fp,vwi->foffset,SEEK_SET);         /* normally foffset=0 */

/* New file, so orbit has to be calculated again */
  if (Find_Window(window,Lab_Main))
  {
    char tmp[100];
    if (strlen(strip_filename(pci->file_name)) +strlen(Lab_Main) <90) 
      sprintf(tmp,"%s    File %s",Lab_Main,strip_filename(pci->file_name));
    else
      sprintf(tmp,"%s    File (name too long!)",Lab_Main);

    gtk_window_set_title((GtkWindow *)Find_Window(window,Lab_Main),tmp);
  }
  pci->orbit.valid=FALSE;

/* Read header. */
  switch(pci->file_type)
  {
#if _SUN_
    case SUN:
      err=read_sun_header(pci);
      if (pci->depth>8)
      {
        err='d';
      }
      vwi->lum_min=0;
      vwi->lum_max=255;
    break;
#endif
#if _PPM_
    case PPM: case PGM:
      err=read_ppm_header(pci);
      vwi->lum_min=0;
      vwi->lum_max=(1<<pci->depth)-1;
    break;
#endif
#if _TIFF_
    case TIFF:
      err=read_tiff_header(pci);
      vwi->lum_min=0;
      vwi->lum_max=255;
    break;
#endif
#if _GIF_
    case GIF:
      err=read_gif_header(pci);
      vwi->lum_min=0;
      vwi->lum_max=255;
    break;
#endif
#if _HRPT1_
    case HRPT1: case CHRPT1: case SEAWIF1:
      err=read_hrpt1_hdr(pci);
      {
        GtkWidget *kepler_window=Find_Window(window,Lab_Keplerwnd);
        if (kepler_window)
          Set_Keplervals(kepler_window,&pci->kepler,&pci->orbit);

        if (window)
          Set_Entry(window,LAB_SATNM,"%s",pci->sat_name);
      }
      vwi->lum_min=0;
      vwi->lum_max=1023;
    break;
#endif
#if _HRI1_
    case HRI1:
      err=read_hri1_hdr(pci);
      if (window)
        Set_Entry(window,LAB_SATNM,"%s",pci->sat_name);
      vwi->lum_min=0;
      vwi->lum_max=255;
    break;
#endif
#if _AHRPT1_
    case AHRPT1:
      err=read_ahrpt1_hdr(pci);
      vwi->lum_min=0;
      vwi->lum_max=1023;
    break;
#endif
    default:
      err='t';
    break;
  }

/* Close file in case of error */
  if (err)
  {
    close_picfile(pci);
    return err;
  }

/* Set zoom factors to 'full' */
  vwi->zval=zoom_full();
  vwi->scan_hor=pci->scan_hor;
  vwi->scan_ver=pci->scan_ver;
  Update_Zoom(window);

/* Set buttons to right state */
  if (window)
  {
/*  Disable window update to prevent multiple redraws */
    Set_Enable_Update(window,FALSE);

    init_menubuts(window,pci,vwi);

/* Activate/deactivate channel selection */
    Channel_Select(window,pci,vwi);
    Update_channel_save(window,pci,vwi);

/* Deselect items in satellite select list */
    Deselect_Clist(window,LAB_WINSELSATLIST,LAB_SELSATLIST);

/*  Enable window update. */
    Set_Enable_Update(window,TRUE);
  }
  else
  {
/* This is to take over width/height into pci. */
    Channel_Select(window,pci,vwi);
  }
  Sense_Button(window,MENU_CLOSE,TRUE);

  return 0;
}

/*************************************
 * Close a file.
 * Set file pointer to zero. This is used to determine 
 *   if a file is opened.
 * pci: struct containing the file pointer.
 *************************************/
void close_picfile(PICINFO *pci)
{
  if (pci->fp) fclose(pci->fp);
  pci->fp=NULL;
}


/*************************************
 * Activate file selection and get file name
 *************************************/
void file_ok_sel(GtkWidget *widget, GtkFileSelection *fs)
{
  GtkWidget *file_window=Find_Parent_Window(widget);
  GtkWidget *main_window=Find_Window(file_window,Lab_Main);
  PICINFO *pci=Get_Picinfo_db(widget);
  VIEWINFO *vwi=Get_Viewinfo_db(widget);
  int err=0;

/* restore file name */
  if (file_selection_get_filename())
  {
    strcpy(pci->file_name,(char *)file_selection_get_filename());
  }

/* Open the selected file. */
  if (!(err=open_file(pci,vwi,main_window)))
  {
    Close_Fileselectf(file_window);

/* Load palette and show new picture */
    load_std_palette(main_window,pci);
    Update_Main(main_window);


  }
  else
  {
/* Error handling. */
    switch(err)
    {
      case 't':
        Create_Message("Error","Unknown file type of \n%s.\n",pci->file_name);
      break;
      case 'o':
        Create_Message("Error","Can't open file.");
      break;
      case 'd':
        Create_Message("Error","Depth=%d currently not supported.\n",pci->depth);
      case 'r':
        Create_Message("Error","read problem in %s.\n",pci->file_name);
      break;
      case 'p':
        Create_Message("Error","No picture found in %s.\n",pci->file_name);
      break;
      default:
        Create_Message("Error","Unknown error %d (%c).\n",err,err);
      break;
    }
  }
}

/*************************************
 * Create file selector
 * Menu: File -> Open
 *************************************/
void MenuSelectFile(GtkWidget *widget)
{
  static char d_filter[50],f_filter[50];
  GtkWidget *main_window=Find_Parent_Window(widget);
  GtkWidget *fwnd;

  PICINFO *pci=Get_Picinfo_db(widget);
  if (!*d_filter) strcpy(d_filter,"*");
  if (!*f_filter) strcpy(f_filter,"*");

  fwnd=Create_Fileselectf(main_window,"Open",NULL,file_ok_sel,
                                        pci->work_location,
                                        d_filter,f_filter,0,NULL);
 if (fwnd)
   place_window(fwnd,prefs.sub_wnd_x,prefs.sub_wnd_y,prefs.sub_wnd_geom);
}

/*************************************
 * Find a name for a file which doesn't exist yet.
 * This name has the form: <prefix>#<suffix> (# = 1 or 2 or 3 ...)
 *************************************/
int create_new_filename(char *new_name,char *loc,char *prefix,char *suffix)
{
  FILE *fp=NULL;
  char tmp[1000];
  int nr=0;
  do
  {
    if (fp) fclose(fp);
    if (nr>1000) return 0;                           /* Can't open file! */
    sprintf(new_name,"%s%02d%s",prefix,++nr,suffix); /* Form new file name. */

    if (loc)
      sprintf(tmp,"%s%s",loc,new_name);              /* new file name tot. path. */
    else
      sprintf(tmp,"%s",new_name);                    /*  */
  } while ((fp=fopen(tmp,"rb")));                    /* Exit if file doesn't exist. */

  return nr;
}

/*************************************
 * Open a new file in write mode.
 * First determine a name for a file which doesn't exist yet.
 * This name has the form: <name>#<ext> (# = 1 or 2 or 3 ...)
 *************************************/
int open_new_recfile(char *name,char *ext,RECORDINFO *rdi)
{
  char file_name[400];
  int nr=0;
  sprintf(file_name,"%s%s",rdi->work_location,name);
  if (!(nr=create_new_filename(rdi->file_name,NULL,file_name,ext))) return 0;

/* Open new file. */
  rdi->fp=fopen(rdi->file_name,"wb");
  return nr;
}

/*************************************
 * Close a rec-file.
 * Set file pointer to zero. This is used to determine 
 *   if a file is opened.
 * rdi: struct containing the file pointer.
 *************************************/
void close_recfile(RECORDINFO *rdi)
{
  if (rdi->fp) fclose(rdi->fp);
  rdi->fp=NULL;
}

/*************************************
 * Close current viewed file.
 * Set file pointer to zero. This is used to determine 
 *   if a file is opened.
 * pci: struct containing the file pointer.
 *************************************/
void close_file(GtkWidget *widget)
{
  GtkWidget *main_window=Find_Parent_Window(widget);
  PICINFO *pci=Get_Picinfo_db(widget);
  close_picfile(pci);
  gtk_window_set_title((GtkWindow *)Find_Window(main_window,Lab_Main),Lab_Main);

  Sense_Button(main_window,MENU_CLOSE,FALSE);
}

