/**************************************************
 * RCSId: $Id$
 *
 * Mapping related functions.
 * Project: WSAT
 * Author: R. Alblas
 *
 * History: 
 * $Log$
 **************************************************/
/*******************************************************************
 * Copyright (C) 2000 R. Alblas. 
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software. If not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 ********************************************************************/

#include <malloc.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>

#include "wsat.h"
#include "orbit.h"
#include "hrpt.h"
#include "menu_wsat.h"

#define MAX_CLRMAP_SIZE 1050
#define DEFAULT_CLRMAP_SIZE 256

/*************************************
 * Read color map loaded in pci into palette
 * Determine color with max lum
 * If pci=NULL -> Generate grey-palette
 *************************************/
void load_std_palette(GtkWidget *window,
                      PICINFO   *pci)
{
  int i;
  int bpp;
  int size_clrmap=DEFAULT_CLRMAP_SIZE;
  GdkColor clr[MAX_CLRMAP_SIZE];
  GdkColormap clrmap;
  GtkWidget *drawing_area=Find_Widget(window,"GTK_DRAWING_AREA");

  if (pci) size_clrmap=MIN(pci->clrmap.size,MAX_CLRMAP_SIZE);
  bpp=(int)(log(size_clrmap+1.)/log(2.));

  clrmap.colors=clr;
  clrmap.size=size_clrmap;

/* Load GdkColor */
  for (i=0; i<clrmap.size; i++)
  {
    if (pci==NULL)
    {
/* No palette from file ==> define grey palette */
      clr[i].red=((i)<<(16-bpp));
      clr[i].green=((i)<<(16-bpp));
      clr[i].blue=((i)<<(16-bpp));
    }
    else
    {
/* Load from PICINFO (so, use colormap defined in file) */
      clr[i].red=pci->clrmap.red[i]<<(16-bpp);
      clr[i].green=pci->clrmap.green[i]<<(16-bpp);
      clr[i].blue=pci->clrmap.blue[i]<<(16-bpp);
    }
  }

/* Load the palette */
  Create_RGBPalette(drawing_area,&clrmap);
}

/*************************************
 * Generate extra palette with colors
 * for zoom etc.
 *************************************/
#define EXTRA_CLRS 8
void load_extra_palette(GtkWidget *window)
{
  int i;
  GdkColor clr[EXTRA_CLRS];
  GdkColormap clrmap;

  clrmap.colors=clr;
  clrmap.size=EXTRA_CLRS;

  for (i=0; i<EXTRA_CLRS; i++)
  {
    clr[i].red  =(i&0x4? 0xffff: 0x0000);
    clr[i].green=(i&0x2? 0xffff: 0x0000);
    clr[i].blue =(i&0x1? 0xffff: 0x0000);
  }
  Create_GCPalette(window,&clrmap);
}

/*************************************
 * Translate x-value to angle position on earth.
 * (angle between line "sat <--> middle point earth" and "pos. earth <--> mp"
 * Needed: sat_height     : height satellite
 *         x              : x-pos. (for HRPT: 0 ... 1023)
 *         xh             : half value of x (1023 for HRPT)
 *         max_sens_angle : max. angle sensor
 * If x=0  ==> alpha=max_sens_angle ==> beta=beta_max
 * if x=xh ==> alpha=0 ==> beta=0        
 *************************************/
float x2earthangle(float sat_height,            /* height sat */
                   SCANDIR scan_hor,            /* scan direction sensor */
                   int x,int xh,                 
                   float max_sens_angle)
{
  float alpha,beta,tmp1,tmp2;

/* angle sensor in satellite */
  alpha=(xh-x)*D2R(max_sens_angle)/xh;
  if (scan_hor==W_E) alpha*=-1;

/* translate into radius angle */
  tmp1=(sat_height+Rearth)/Rearth;
  tmp2=(tmp1*tmp1*sin(alpha)*sin(alpha));
  beta=acos(tmp1*sin(alpha)*sin(alpha)+
                 cos(alpha)*sqrt(1-tmp2));
  if (alpha<0) beta*=-1;
  return beta;
}

#define EXPAND_L 2
/*************************************
 * Generate table to linearize HRPT
 * Needed: sat_height     : height satellite
 *         max_sens_angle : max. angle sensor
 * Return: pointer to table; table[x]=actual_x
 *************************************/
int *mk_lintbl(PICINFO  *pci,
               VIEWINFO *vwi)
{
  int *xtbl,*xrevtbl,*lintbl;

  int x;
  float max_sens_angle=pci->max_sens_angle; /* angle */
  float height=pci->orbit.height;
  float beta,beta_max;

  int pox=pci->ox;
  int width=pci->width;

  int i_width=pci->width_original*EXPAND_L;
  int i_hwidth=i_width/2;

  if (!vwi->linearize) return NULL;
  if (!pci->width_original) return NULL;
/* If no height is specified: Make it 800 km. */
  if (!height) height=800000;

  if ((lintbl=(int *)calloc(i_width,sizeof(*lintbl)))==NULL) return NULL;
  if ((xtbl=(int *)calloc(i_width,sizeof(*xtbl)))==NULL) return NULL;
  if ((xrevtbl=(int *)calloc(width,sizeof(*xtbl)))==NULL) return NULL;

/* Calculate max. angle projected on earth surface. */
  beta_max=x2earthangle(height,0,0,1,max_sens_angle);

/*
  Create lin. table.
  First expand, to prevent loss of pixels:
    0 1 2 3 4... ==> 0 0 1 1 2 2 3 3 4 4
  Then calulate and load x-mapping in lintbl:
    0 0 0 0 1 1 1 1 ... 1019 1020 1021 1022 1023 ... 2047 2047 2047 2047 
*/
  {
    int lxp=0,ilx,lx;

    lintbl[0]=0;
    lintbl[i_width-1]=pci->width_original-1;

/* calculate for half width; second half is same */
    for (x=0; x<i_hwidth; x++)
    {
/* Angle point_1024 -> center_earth -> point_x */
      beta=x2earthangle(height,0,x,i_hwidth,max_sens_angle);

/* lx runs from 0 (beta=beta_max) to 2047 (beta=0) */ 
      lx=(i_hwidth-1)-(int)((i_hwidth-1)*beta/beta_max);
      for (ilx=lxp; ilx<=lx; ilx++)
      {
        lintbl[ilx]=x/EXPAND_L;
        lintbl[i_width-1-ilx]=pci->width_original-1-lintbl[ilx];
      }
      lxp=lx+1;
    }
  }

  if (pox+width > pci->width_original) pox=pci->width_original-width; 
  pox*=EXPAND_L;

/* Load x-mapping. Picture-offset 'pci->ox' is taken into account here. 
   Mapping to xtbl is done such that xtbl[0]=0 and xtbl[width-1]=width-1;
   So, some resolution in the middle gets lost!
   Load inversed-table too.
*/
  for (x=0; x<i_width; x++)
  {
    int xlin;
    if (x+pox > ((pci->width_original*EXPAND_L)-1)) break;

    xlin=((long)(lintbl[x+pox]-lintbl[pox])*(long)(width-1)) /
                        (long)(lintbl[EXPAND_L*(width-1)+pox]-lintbl[pox]);

    xtbl[x]=xlin;
    if ((xlin>=0) && (xlin<width)) xrevtbl[xlin]=x;
  }

  free(lintbl);

/* Replace old tables with new ones */
  if (vwi->lin_xtbl) { free(vwi->lin_xtbl); vwi->lin_xtbl=NULL; }
  if (vwi->revlin_xtbl) { free(vwi->revlin_xtbl); vwi->revlin_xtbl=NULL; }

  vwi->lin_xtbl=xtbl;
  vwi->revlin_xtbl=xrevtbl;

  return xtbl;
}


/*************************************
 * Define x-y mapping of picture to window
 * xmap contains X-mapping (xmap[window-x-val]=file-line-position)
 * ymap contains Y-mapping
 *************************************/
int map_xy(int      wnd_width,
           int      wnd_height,
           gboolean single_channel,
           PICINFO  *pci,
           VIEWINFO *vwi,
           guint    **xmap,
           guint    **ymap)

{
  int i;
  int pos;
  int inter_channel,header_len,chan_sel;

  mk_lintbl(pci,vwi);

  if ((*xmap=(guint *)malloc(wnd_width*sizeof(guint)))==NULL) return 1;
  if ((*ymap=(guint *)malloc(wnd_height*sizeof(guint)))==NULL) return 2;

/* dx/dy=# screen-pixel used to show 1 actual pixels */
  vwi->zval.dx=vwi->zval.zx*(float)wnd_width/(float)pci->width;
  vwi->zval.dy=vwi->zval.zy*(float)wnd_height/(float)pci->height;

/* If one channel, or all channels of multi-channel file needed: */
  if (single_channel)
  {
    chan_sel=0;
    inter_channel=1;
    header_len=0;
  }
  else
  {
/* Determine which channel is selected */
    chan_sel=0;
    for (i=1; i<NR_CHANNELS_MAX; i++)
    {
      if ((vwi->channel_selected[i]) && (pci->channels_recorded[i])) 
        break;
      if (pci->channels_recorded[i]) chan_sel++;
    }
    inter_channel=pci->fri.inter_chan;
    header_len=pci->fri.header_len;
  }

/* Calculate mapping x-direction */
  for (i=0; i<wnd_width; i++)
  {
    pos=scrx2picx(i,vwi,pci)-pci->ox;
    pos=pos*inter_channel;     /* take into account channel interleaf */
    pos+=chan_sel;             /* select channel */
    pos+=header_len;  /* pic starts at line[pci->fri.header_len] */
    *(*xmap+i)=pos;
  }

/* Calculate mapping y-direction */
  for (i=0; i<wnd_height; i++)
  {
    pos=scry2picy(i,vwi,pci)-pci->oy;
    pos=MIN(pos,pci->height-1);
    pos=MAX(pos,0);
    *(*ymap+i)=pos;
  }
  return 0;
}

void scroll_update(GtkWidget *window,PICINFO *pci,VIEWINFO *vwi)
{
  int valx,valy;

  if (vwi->scan_hor==E_W)
    valx=pci->width*(1.-1./vwi->zval.zx)-vwi->zval.ox;
  else
    valx=vwi->zval.ox;

  if (vwi->scan_ver==S_N)
    valy=pci->height*(1.-1./vwi->zval.zy)-vwi->zval.oy;
  else
    valy=vwi->zval.oy;

  Set_Scroll(window,"hbar",pci->width,&valx,&vwi->zval.zx,
                    "vbar",pci->height,&valy,&vwi->zval.zy);
}

/*************************************
 * Define mapping of picture to window
 * xmap contains X-mapping
 * ymap contains Y-mapping
 * zmap contains color-mapping
 *************************************/
int map_pic(GtkWidget *window,
            PICINFO   *pci,
            VIEWINFO  *vwi,
            gboolean  single_channel,
            guint     **xmap,
            guint     **ymap,
            guint     **zmap)
{
  GtkWidget *drawing_area=Find_Widget(window,"GTK_DRAWING_AREA");
  RGBI *rgbi=Get_RGBI(drawing_area);
  int clrmap_size;
  int i;
  int lval;
  int wnd_width;           /* width window */
  int wnd_height;          /* height window */
  int zmax;                /* colormap size picture */
  float lmin,lmax;
  if (!window) return -1;

  clrmap_size=rgbi->size; 
  wnd_width=drawing_area->allocation.width;   /* width window */
  wnd_height=drawing_area->allocation.height; /* height window */
  zmax=pci->clrmap.size;                      /* colormap size picture */
 
  if ((map_xy(wnd_width,wnd_height,single_channel,pci,vwi,xmap,ymap))) return 1;

/* Calculate mapping z-direction */

  if ((*zmap=(guint *)malloc(zmax*sizeof(guint)))==NULL) return 3;
  if (vwi->lum_max==vwi->lum_min) vwi->lum_max++;
  lmin=vwi->lum_min;
  lmax=vwi->lum_max;

  vwi->gamma=Get_Adjust(window,LAB_GAMMAVAL);

  if ((vwi->gamma) && (vwi->gamma!=1.))
  {
    lmin=pow((float)vwi->lum_min,1./vwi->gamma);
    lmax=pow((float)vwi->lum_max,1./vwi->gamma);
  }


  for (i=0; i<zmax; i++)
  {
    if (i<vwi->lum_min)
      lval=0;
    else if (i>vwi->lum_max)
      lval=clrmap_size-1;
    else if ((vwi->gamma) && (vwi->gamma!=1.))
    {
      float gi;
      gi=pow((float)i,1./vwi->gamma);
      lval=(clrmap_size*(gi-lmin))/(lmax-lmin);
    }
    else
      lval=(clrmap_size*(i-vwi->lum_min))/(vwi->lum_max-vwi->lum_min);

    lval=MAX(lval,0); lval=MIN(lval,clrmap_size-1);
    if (vwi->invert)
      *(*zmap+i)=clrmap_size-lval;
    else
      *(*zmap+i)=lval;
  }

  scroll_update(window,pci,vwi);

  return 0;
}


/*************************************
 * Translate screen x into picture x 
 *************************************/
int scrx2picx(int scrx,VIEWINFO *vwi,PICINFO *pci)
{
  float x;
  if (vwi->scan_hor==E_W)
  {
    x=vwi->zval.ox-(float)scrx/vwi->zval.dx+pci->width/vwi->zval.zx; 
  }
  else
  {
    x=vwi->zval.ox+(float)scrx/vwi->zval.dx;
  }

  x=MIN(x,vwi->zval.ox+pci->width-1); /* clip pos max */

  if (vwi->linearize)
  {
    if (!vwi->lin_xtbl) mk_lintbl(pci,vwi);
    if (vwi->lin_xtbl) x=vwi->lin_xtbl[(int)(x*EXPAND_L)];
  }

  x+=pci->ox;
  return (int)x;
}

/*************************************
 * Translate picture x into screen x 
 *************************************/
int picx2scrx(int x,VIEWINFO *vwi,PICINFO *pci)
{
  int scrx;
  x-=pci->ox;
  if (vwi->linearize)
  {
    if (!vwi->revlin_xtbl) mk_lintbl(pci,vwi);
    if (vwi->revlin_xtbl) x=vwi->revlin_xtbl[x]/EXPAND_L;
  }
  if (vwi->scan_hor==E_W)
  {
    scrx=(pci->width/vwi->zval.zx-(x-vwi->zval.ox))*vwi->zval.dx;
  }
  else
  {
    scrx=(x-vwi->zval.ox)*vwi->zval.dx;
  }

  return scrx;
}

/*************************************
 * Translate screen y into picture y 
 *************************************/
int scry2picy(int scry,VIEWINFO *vwi,PICINFO *pci)
{
  int y;
  if (vwi->scan_ver==S_N)
  {
    y=vwi->zval.oy-(float)scry/vwi->zval.dy+pci->height/vwi->zval.zy;
  }
  else
  {
    y=vwi->zval.oy+(float)scry/vwi->zval.dy;
  }

  y+=pci->oy;
  return y;
}

/*************************************
 * Translate picture y into screen y 
 *************************************/
int picy2scry(int y,VIEWINFO *vwi,PICINFO *pci)
{
  int scry;
  y-=pci->oy;
  if (vwi->scan_ver==S_N)
  {
    scry=(pci->height/vwi->zval.zy-(y-vwi->zval.oy))*vwi->zval.dy;
  }
  else
  {
    scry=(y-vwi->zval.oy)*vwi->zval.dy;
  }
  return scry;
}


#define MINTEMP (-70+273)
#define MAXTEMP (+70+273)
/*************************************
 * Map temperature on 'rainbow color-map'. 
 * Input is temp. in .1 K.
 *************************************/
gint map_temp(gint c,                           /* temperature in Kelvin */
              guint *zmap,                      
              guint16 *r,guint16 *g,guint16 *b) /* result red/green/blue */
{
  gint c1;

/* Convert temp into Kelvin. */
  c/=10;

/* Convert temp. range [MINTEMP ... MAXTEMP] of 'c' into [0-1000]  */ 
  c=MIN(MAX(c,MINTEMP),MAXTEMP);        /* clip into range [MINTEMP-MAXTEMP] */
  c=(c-MINTEMP)*1023/(MAXTEMP-MINTEMP); /* convert to [0-1023] */

  c=MIN(MAX(c,0),1023);                 /* clip to [0-1023] */
  c1=zmap[c];                           /* Translate to screen value */
/*  c1=MIN(c1,245);*/

/* Translate into 'rainbow' */
  *g=127.*(1+sin(D2R(c1)));
  *b=127.*(1+sin(D2R(c1+120)));
  *r=127.*(1+sin(D2R(c1+240)));

  return c;
}

/*************************************
 * Translate 'rgb' into temperature
 *************************************/
gint inv_maptemp(GdkColor clr,VIEWINFO *vwi,int size)
{
  gint z;
/* Translate (r,g,b) into z (inverse 'rainbow' translation) */
  if ((clr.blue>100) && (clr.red<50))
    z=R2D(asin(clr.green/127.-1.));
  else if (clr.red<225)
    z=R2D(asin(clr.red/127.-1.))+120;
  else
    z=R2D(asin(clr.blue/127.-1.))+240;

/* Take into account zmap */
  z=z*(vwi->lum_max-vwi->lum_min)/size+vwi->lum_min;

/* Translate to temp. in Celcius */
  z=z*(MAXTEMP-MINTEMP)/1000+MINTEMP-273;
  return z;
}

