/**************************************************
 * RCSId: $Id$
 *
 * Common parallel port functions
 * Project: WSAT
 * Author: R. Alblas
 *
 * Note: This file requires the -O option during compilation
 *       because of the inw/outb etc. functions.
 * History: 
 * $Log$
 **************************************************/
/*******************************************************************
 * Copyright (C) 2000 R. Alblas. 
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software. If not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 ********************************************************************/
#include <asm/io.h>
#include <unistd.h>
#include "pp_dec_interf.h"
#include "wsat.h"

#if __GTK_WIN32__
/* Define ioperm for Windows; doesn't do anything. 
  (Should do something for Windows XP??)
*/
int ioperm(int a,int b,int c)
{
  return 0;
}
#endif

/*************************************
 * Get data from decoder.
 * Decoder is connected to parallel port.
 * Return: TRUE  if succeeded
 *         FALSE if time-out reached without ready-signal
 *************************************/
#define TIME_OUT 10000
gboolean get_pp_data(RECORDINFO *rdi,int *datstat)
{
  int addr_portd=rdi->parport_base;
  int addr_portc=rdi->parport_base+2;
  gboolean sel_hw;
  int n=0;

  sel_hw=rdi->sel_hw;
  do
  {
    GET_DS_PP(*datstat,addr_portd);
    if (n++ > TIME_OUT)
    {
      return FALSE;                        /* No signal! */
    }
  } while ((READYN_PP(*datstat)));

/* Got readyn; read par. port again to get stable data.
   If fast_parportaccess flag is set don't read;
   in this case readyn should be hardware delayed to get proper data! 
*/

  if (!rdi->fast_parportaccess)
    GET_DS_PP(*datstat,addr_portd);

/* Generate strobe so decoder can send next data */
  STROBE_PP(sel_hw,addr_portc);

  return TRUE;
}

/*************************************
 * Wait for in-sync of decoder.
 * Decoder is connected to parallel port.
 * Return: True if sync-start found
 *************************************/
gboolean pp_wait_for_insync(RECORDINFO *rdi,int threshold)
{
  int addr_portd=rdi->parport_base;
  int datstat;
  guint32 count=0;
  guint32 threshold_word;
  int n=0;

  if (ioperm(addr_portd,3,1)) return FALSE;

/* Translate #frames-to-wait to #words-to-wait */
  threshold_word=threshold*rdi->fri.words_per_frame;
  do
  {
    if (!rdi->recording) return FALSE;           /* Stop if recording aborted */
    if ((get_pp_data(rdi,&datstat)))                /* Get data */
    {
      if (INSYNC_PP(datstat))
        count++;
      else
        count=0;
    }
    else
    {
/*
   waited too long for new data; problem with readyn? 
   Try again,but first do g_main_iteration to catch e.g. 'Stop' event.
*/
    }

/* Evaluate gui-events now and then. (Not too often.) */
    if (n++>100)
    {
      while (g_main_iteration(FALSE));
      n=0;
    }
  } while (count<threshold_word);

  return TRUE;
}

/*************************************
 * Wait for out-sync of decoder.
 * Decoder is connected to parallel port.
 * Return: True if sync-start found
 *************************************/
gboolean pp_wait_for_outsync(RECORDINFO *rdi,int threshold)
{
  int addr_portd=rdi->parport_base;
  int datstat;
  guint32 count=0;
  guint32 threshold_word;
  int n=0;

  if (ioperm(addr_portd,3,1)) return FALSE;

/* Translate #frames-to-wait to #words-to-wait */
  threshold_word=threshold*rdi->fri.words_per_frame;
  do
  {
    if (!rdi->recording) return FALSE;           /* Stop if recording aborted */
    if ((get_pp_data(rdi,&datstat)))                /* Get data */
    {
      if (!INSYNC_PP(datstat))
        count++;
      else
        count=0;
    }
    else
    {
/*
   waited too long for new data; problem with readyn? 
   Try again,but first do g_main_iteration to catch e.g. 'Stop' event.
*/
    }

/* Evaluate gui-events now and then. (Not too often.) */
    if (n++>1000)
    {
      while (g_main_iteration(FALSE));
      n=0;
    }
  } while (count<threshold_word);

  return TRUE;
}

/*************************************
 * Select channel on decoder.
 * Decoder is connected to parallel port.
 *************************************/
void select_channel_pp(gboolean   *chan,
                       RECORDINFO *rdi,
                       gboolean    reprogramming_fpga)
{
  int rego=0;
  int addr_portd=rdi->parport_base;
  int addr_portc=addr_portd+2;

/* Channel select */
  if (chan)
  {
    if (chan[0]) rego|=PP_FRHDR_SEL;
    if (chan[1]) rego|=PP_CHAN1_SEL;
    if (chan[2]) rego|=PP_CHAN2_SEL;
    if (chan[3]) rego|=PP_CHAN3_SEL;
    if (chan[4]) rego|=PP_CHAN4_SEL;
    if (chan[5]) rego|=PP_CHAN5_SEL;
  }
  SEL_CHAN_PP(rego,rdi->sel_hw,addr_portd,addr_portc);

/* If reprogramming FPGA, wait for 1 second and write selection again */
  if (reprogramming_fpga)
  {
    sleep_ms(1000);
    SEL_CHAN_PP(rego,rdi->sel_hw,addr_portd,addr_portc);
  }
}

int get_ppreg(int pp_addr)
{
  return inb(pp_addr);
}

int do_batch(RECORDINFO *rdi,PICINFO *pci,VIEWINFO *vwi)
{
  char str[20],*p;
  gboolean chan[6];
  int i;
  strcpy(str,rdi->rec_chan);
                                                                    
  if ((p=strchr(str,':'))) { *p=0; p++; }
  if ((p) && (strchr(p,'g'))) rdi->gen1_dec0=TRUE;
  if ((p) && (strchr(p,'a'))) p="0123456789";
                                                                    
  if (ioperm(rdi->parport_base,3,1))
  {
    log_message("Error: Can't access Parallel port with address 0x%x.",rdi->parport_base);
    return 1;
  }
  if (!rdi->hwsel_type)
    rdi->sel_hw=rdi->gen1_dec0;

  if (!strncmp(str,"chrpt",5)) { rdi->file_type=CHRPT; rdi->sat_type=CHRPT; }
  if (!strncmp(str,"hrpt",5)) { rdi->file_type=HRPT; rdi->sat_type=HRPT; }
 
  chan[0]=TRUE;
  rdi->channels_recording[0]=TRUE;
  for (i=1; i<=5; i++)
  {
    chan[i]=FALSE;
    switch (rdi->file_type)
    {
      case HRPT:
        if (strchr(p,i+'0'))
        {
          chan[i]=TRUE;
          rdi->channels_recording[i]=TRUE;
        }
      break;
      case CHRPT:
        if (i==0)
        {
          chan[0]=TRUE;
          rdi->channels_recording[0]=TRUE;
        }
        else if ((strchr(p,(i*2-1)+'0')) || (strchr(p,(i*2)+'0')))
        {
          chan[i]=TRUE;
          rdi->channels_recording[i*2-1]=TRUE;
          rdi->channels_recording[i*2]=TRUE;
        }
      break;
      default:
      break;
    }
  }

  select_channel_pp(chan,rdi,TRUE);
  rdi->recording=TRUE;
  rdi->save_flag=TRUE;
  if (!strcmp(str,"hri"))
  {
    Recordpp_Hri(NULL,rdi,pci,vwi);
  }
  else if (!strcmp(str,"hrpt"))
  {
    rdi->file_type=HRPT1;
    rdi->sat_type=HRPT;
    Recordpp_Hrpt(NULL,NULL,rdi,pci,vwi);
  }
  else if (!strcmp(str,"chrpt"))
  {
    rdi->file_type=CHRPT1;
    rdi->sat_type=CHRPT;
    Recordpp_Hrpt(NULL,NULL,rdi,pci,vwi);
  }
  return 0;
}
