/**************************************************
 * RCSId: $Id$
 *
 * Menu/gui for Loading new Kepler data.
 * Project: WSAT
 * Author: R. Alblas
 *
 * History: 
 * $Log$
 **************************************************/
/*******************************************************************
 * Copyright (C) 2000 R. Alblas. 
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software. If not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 ********************************************************************/
#include <string.h>
#include <malloc.h>
#include "wsat.h"
#include "orbit.h"

extern PREFS prefs;

/*************************************
 * Search satellite in NORAD 2 lines format file
 *
 * fp            : file pointer to opened file 
 * satname_tofind: name of satellite to find
 * kepler        : result Kepler data
 *
 * Return: 1 if succeeded
 *         0 if failed (EOF)
 *************************************/
int get_satkeps(FILE *fp,char *satname_tofind,KEPLER *kepler)
{
  char satname_infile[100];

/* Reset Kepler-vars */
  kepler->epoch_year=0;
  kepler->epoch_day=0.;
  kepler->decay_rate=0.;
  kepler->inclination=0.;
  kepler->raan=0.;
  kepler->eccentricity=0.;
  kepler->perigee=0.;
  kepler->anomaly=0.;
  kepler->motion=0.;
  kepler->epoch_rev=0;

  if (!fp) return 0;

/* Search in file requested satellite name. */
  rewind(fp);
  while ((read_norad_next_keps(fp,satname_infile,kepler)))
  {
    if (!strcmp(satname_infile,satname_tofind))
      return 1;
  }
/* Satellite not found! */
  return 0;
}


/* Next global var has 'file' scope.
   Needed to connect file pointer in Create_selectsat_window
   to Selection_made part */
static FILE *fp_glbl;

#define LAB_Diftm "Diff start & Kepler ref. time (days)"

/*************************************
 * Function activated as soon as a line 
 * in the clist is selected.
 *************************************/
void Selection_made( GtkWidget      *clist,
                     gint            row,
                     gint            column,
		     GdkEventButton *event,
                     gpointer        data)
{
  GtkWidget *main_window=Find_Parent_Window(clist);
  PICINFO *pci;
  gchar *text;
  KEPLER kepler;
  ORBIT orbit;

  pci=Get_Picinfo_db(clist);

  if (!pci)
  {
    Create_Message("Software error","Selection_made: clist=%x\n",(long)clist);
    return;
  }

/* Get line selected. Contains sat-name (in column 0). */
  gtk_clist_get_text(GTK_CLIST(clist), row, 0, &text);

  if (strlen(text)>90)
  {
    Create_Message("Problem","Selection_made: length satname=%d\n",
                                                          strlen(text));
    return;
  }

/* Load Kepler-data of selected sat. */
  if (!(get_satkeps(fp_glbl,text,&kepler)))
  {
    Create_Message("Error","Can't find %s.\n%s\n%s",text,
                   "NOTE: This error should never pop up.",
                   "Software error??");
    return;
  }

/* Copy extracted kepler-values */
  pci->kepler=kepler;

/* If Kepler-edit windo is active then update it */
  Set_Keplervals(main_window,&pci->kepler,NULL);

/* Calculate some orbit params */
  calc_orbitconst(&kepler,&orbit);

/* recalculate time difference start-time and satelite reference time (days) */
  {
    gint32 t_strt,t_sref;
    t_strt=mktime_ntz(&pci->orbit.start_tm_comp)/3600/24;
    t_sref=mktime_ntz(&orbit.ref_tm)/3600/24;
    Set_Entry(main_window,LAB_Diftm,"%d",t_strt-t_sref);
  }
}

/*************************************
 * Function activated as soon as a line 
 * in the clist is unselected.
 *************************************/
void Selection_lost( GtkWidget      *clist,
                     gint            row,
                     gint            column,
		     GdkEventButton *event,
                     gpointer        data)
{
  GtkWidget *main_window=Find_Parent_Window(clist);
  PICINFO *pci=Get_Picinfo_db(clist);
  if (!pci) return;

/* Reload Kepler data from file if no sat is selected */
  Reload_Kepler(pci);

/* recalculate time difference start-time and satelite reference time (days) */
  {
    gint32 t_strt=mktime_ntz(&pci->orbit.start_tm_comp)/3600/24;
    gint32 t_sref=mktime_ntz(&pci->orbit.ref_tm)/3600/24;
    Set_Entry(main_window,LAB_Diftm,"%d",t_strt-t_sref);
  }
}

/*************************************
 * Managing buttons in Create_selectsat_window.
 *************************************/
void select_sat_func(GtkWidget *widget, gpointer data)
{
  GtkWidget *select_window=gtk_widget_get_toplevel(widget);
  PICINFO *pci=Get_Picinfo_db(widget);
  char *name=(char *)data;

/* Activate advanced orbit window */
  if (!strcmp(name,LAB_Advwnd))
  {
    Create_kepleredt_wnd(select_window,&pci->kepler,NULL);
  }

/* Redraw picture with new Kepler-data */
  if (!strcmp(name,LAB_Redraw))
  {
    Update_Main(select_window);
  }

/* Save selected Kepler data into picture file */
  if (!strcmp(name,LAB_Save))
  {
    Rewrite_hdr(pci);
  }

/* Exit kepler-window */
  if (!strcmp(name,LAB_Exit))
  {
    Close_Window(select_window);
  }
}

/*************************************
 * Collect all available sats and select
 *************************************/
void Create_selectsat_window(GtkWidget *main_window,FILE *fp)
{
  KEPLER kepler;
  GtkWidget *wnd,*w1,*w2,*w3,*w4,*wa,*wb,*wy,*wz;
  char sat_name[100];
  int n=0;
  fp_glbl=fp;
/* create the window */
  if (!(wnd=Create_Window(main_window,240,200,LAB_WINSELSATLIST,NULL))) 
    return;

/* Create the list (will be filled lateron) */
  wa=Create_Clist(LAB_SELSATLIST,Selection_made,Selection_lost,2,
                                   "Satellite",12,"Ref time",16,0);

/* Show difference start-time and reference time */
  wb=Create_Entry(LAB_Diftm,NULL,"      ");

  w1=Create_Button(LAB_Advwnd,select_sat_func);
  w2=Create_Button(LAB_Redraw,select_sat_func);
  w3=Create_Button(LAB_Save,select_sat_func);
  w4=Create_Button(LAB_Exit,select_sat_func);

  wy=Pack("",'h',w1,10,w2,10,w3,10,w4,10,NULL);
  wz=Pack(NULL,'v',wa,5,wb,5,wy,5,NULL);
  gtk_container_add(GTK_CONTAINER(wnd),wz);
  gtk_widget_show(wnd);
  place_window(wnd,prefs.sub_wnd_x,prefs.sub_wnd_y,prefs.sub_wnd_geom);

/* All buttons ready. Now fill the list. */
  rewind(fp);
  while (read_norad_next_keps(fp,sat_name,&kepler))
  {
    char date[100];
    char *tmp[4];
    n++;
/* For 'safety' reasons: Stop at a ridiculous amount */
    if (n>1000)
    {
      tmp[0]="More...";
      tmp[1]="(truncated at 1000)";
      gtk_clist_append(GTK_CLIST(wa), tmp);
      break;
    }
    sprintf(date,"Day %d year %d",(int)kepler.epoch_day,kepler.epoch_year);

/* column is sat-name, column 2 shows reference time */
    tmp[0]=sat_name;
    tmp[1]=date;
    gtk_clist_append(GTK_CLIST(wa), tmp);
  }


/* Check if Kepler data is found */
  if (n==0)
  {
    Create_Message("Warning","Kepler-file empty.");
    Close_Window(wnd);
  }
}


/*************************************
 * Activate file selection and get kepler file name
 *************************************/
void kepfile_ok_sel(GtkWidget *widget, GtkFileSelection *fs)
{
  GtkWidget *file_window=Find_Parent_Window(widget);
  GtkWidget *main_window=Find_Window(file_window,Lab_Main);
  FILE *fp;
  char file_name[100];

  if (file_selection_get_filename())
  {
    strcpy(file_name,(char *)file_selection_get_filename());
  }

/* Open the selected file. */
  if ((fp=fopen(file_name,"r")))
  {
    Create_selectsat_window(main_window,fp);
/* Close the file select window */
    Remove_Window_Info(file_window);
    gtk_widget_destroy(file_window); 
  }
  else
  {
    Create_Message("Error","Can't open %s.\n",file_name);
  }
}

/*************************************
 * Create file selector for loading Kepler data
 * Activated by menu: File->Load kepler
 *************************************/
void Menu_LDKepler(GtkWidget *widget)
{
  GtkWidget *main_window=Find_Parent_Window(widget);
  PICINFO *pci=Get_Picinfo_db(widget);
  GtkWidget *fwnd;
  fwnd=Create_Fileselectf(main_window,"Open Kepler-file",NULL,kepfile_ok_sel,
                                        pci->kep_location,NULL,NULL,0,NULL);
  place_window(fwnd,prefs.sub_wnd_x,prefs.sub_wnd_y,prefs.sub_wnd_geom);
}
