/**************************************************
 * RCSId: $Id$
 *
 * SUN Raster related routines
 * Project: WSAT
 * Author: R. Alblas
 *
 * History: 
 * $Log$
 **************************************************/
/*******************************************************************
 * Copyright (C) 2000 R. Alblas. 
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software. If not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 ********************************************************************/

#include <stdio.h>
#include "sun.h"
#include "wsat.h"

/*************************************
 * Read SUN Raster format header
   into PICINFO struct.
 * Return value:
 *   0 if OK
 *   !0 if error
 *************************************/

int read_sun_header(PICINFO *pci)
{
  FILE *fp=pci->fp;
  struct rasterfile hs;
  char clrmap[256];
  int i;

  rewind(fp);

/* Read header */

  if (fread(&hs,1,sizeof(hs),fp)!=sizeof(hs)) return 1;

/* Check sign: Is this really SUN format? */
  if (GUINT32_FROM_BE(hs.ras_magic)!=SUNSIGN) return 2;

/* Copy size info into picinfo structure */
  pci->width=GUINT32_FROM_BE(hs.ras_width);
  pci->height=GUINT32_FROM_BE(hs.ras_height);
  pci->depth=GUINT32_FROM_BE(hs.ras_depth);
  pci->fri.bytes_per_line=GUINT32_FROM_BE(hs.ras_width);
  pci->fri.words_per_line=GUINT32_FROM_BE(hs.ras_width);
  for (i=0; i<NR_CHANNELS_MAX; i++)
    pci->channels_recorded[i]=FALSE;
  pci->channels_recorded[1]=TRUE;
  pci->fri.inter_chan=1;

/* Copy color map into picinfo structure */
  pci->clrmap.size=GUINT32_FROM_BE(hs.ras_maplength)/3;
/* Max. color size=256 */
  if (pci->clrmap.size>256) return 3;

/* read color map */
  fread(clrmap,pci->clrmap.size,1,fp);
  for (i=0; i<pci->clrmap.size; i++) pci->clrmap.red[i]=clrmap[i];
  fread(clrmap,pci->clrmap.size,1,fp);
  for (i=0; i<pci->clrmap.size; i++) pci->clrmap.green[i]=clrmap[i];
  fread(clrmap,pci->clrmap.size,1,fp);
  for (i=0; i<pci->clrmap.size; i++) pci->clrmap.blue[i]=clrmap[i];

/* Start of picture here */
  pci->pic_strt=sizeof(hs)+GUINT32_FROM_BE(hs.ras_maplength);
  pci->fri.header_len=0;

  pci->scan_hor=W_E;
  pci->scan_ver=N_S;

  return 0;
}

/*************************************
 * Write SUN Raster format header
   from PICINFO struct.
 * Return value:
 *   0 if OK
 *   !0 if error
 *************************************/
int write_sun_header(PICINFO *pci)
{
  struct rasterfile hs;
  char clrmap[256];
  int clrmap_size;
  int i;
  int depth;
  FILE *fp=pci->fp;

  clrmap_size=MIN(pci->clrmap.size,256);
/* Depth: only 1/8/24/32 depth are valid? (xv)
   maplen may be < 2**depth
*/
  depth=8;

  hs.ras_magic=GUINT32_TO_BE(SUNSIGN);
  hs.ras_width=GUINT32_TO_BE(pci->width);
  hs.ras_height=GUINT32_TO_BE(pci->height);
  hs.ras_depth=GUINT32_TO_BE(depth);
  hs.ras_length=GUINT32_TO_BE(pci->width*pci->height);
  hs.ras_type=GUINT32_TO_BE(1);
  hs.ras_maptype=GUINT32_TO_BE(1);
  hs.ras_maplength=GUINT32_TO_BE(clrmap_size*3);

/* Write header */
  if (fwrite(&hs,1,sizeof(hs),fp)!=sizeof(hs)) return 1;

/* Copy color map from picinfo structure */
  for (i=0; i<clrmap_size; i++) clrmap[i]=pci->clrmap.red[i];
  fwrite(clrmap,clrmap_size,1,fp);
  for (i=0; i<clrmap_size; i++) clrmap[i]=pci->clrmap.green[i];
  fwrite(clrmap,clrmap_size,1,fp);
  for (i=0; i<clrmap_size; i++) clrmap[i]=pci->clrmap.blue[i];
  fwrite(clrmap,clrmap_size,1,fp);

  return 0;
}

/*************************************
 * Read line 'y' 
 * Return value:
 *   Amount of read bytes (=width picture)
 *************************************/
int read_sun_line(PICINFO *pci,int y,guint16 *line)
{
  FILE *fp=pci->fp;
  guchar pix;
  int ret=0,i;
  fseek(fp,pci->pic_strt + y*pci->fri.bytes_per_line,SEEK_SET);
  for (i=0; i<pci->fri.bytes_per_line; i++)
  {
    ret+=fread(&pix,1,1,fp);
    *(line+i)=pix;
    if ((*(line+i)) > 256) fprintf(stderr,"!!!read_sun_line: %d\n",(*(line+i)));
  }
  return ret;
}

/*************************************
 * Write line 
 * Return value:
 *   Amount of written bytes (=width picture)
 *************************************/
int write_sun_line(PICINFO *pci,guint16 *line)
{
  FILE *fp=pci->fp;
  int width=pci->fri.bytes_per_line;
  char pix;
  int ret=0,i;
  for (i=0; i<width; i++)
  {
    pix=*(line+i);
    ret+=fwrite(&pix,1,1,fp);
  }
  return ret;
}
